// javascript "watch" library for carCosts

/*
Copyright 2013. Daniel Hellerstein (danielh@crosslink.net)

This file is part of carCosts2.

    carCosts2 is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    carCosts2 is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with carCosts2.  If not, see <http://www.gnu.org/licenses/>.


A.    What does this do
B.    Terminology and basics
C.    Initialization
D.    Implementation
D.i.         Where to put these attributes?
D.ii.        Specifying a watch value
D.iii.       Priority
D.iv.        Syntax of watchCall functions
D.v.         Returns from watchCall functions
D.v.1.       Simple example of custom function:
E.    Features
E.i:       Alerts and confirms
E.ii:      Number checking
E.iii      EVAL : evaluating simple expressions
E.iv:      Flags
E.v:       Updates
E.vi:      Saving parameters
F.    Special tags
F.i:       <alert>
F.ii:      <help>
F.iii:     <more>
G.    Useful functions
G.i        Built in functions that can be used in watchCall
G.i.a      Built in functions that can be used in watchCall: details
H.    Notes and reminders:

                   -------------------------

A.  What does this do

     The functions presented here allow one to set up a complex "watch" architecture using simple attributes in existing html tags.

      Elements can "watch" other elements. If anyone of these "watched" elements changes (and this could be the contents of <SPAN>, or the value of an <input>), then
      an element-specific list of functions is called.

      Thus,  consider some element X (say, an <input> and elements Y1 and Y2 whose value depends on the current value of X
      One could have onchange events associated with X that update Y1 and Y2.
      Instead, this "watch" facility tells Y1 and Y2 to "watch for changes on X", and if they occur -- update your values.

      And this is recursive: elements that are updated because an element they "watch" has changed will result in other elements (that watch the updated element) also being updated.

      IOW: the responsiblity for updating is assigned to the element to be updated, rather than the elements that might change.
           This makes it easier to add and modify elements, you don't have to be as careful with onchange events!

B.  Terminology and basics

      This description will refer to "changedElement" and "thisElement". These are used verbatim as argument names, and values of options.
      More importantly, in a general sense they refer to elements that are "being watched" and are "doing the watching".

      The basic idea is one identifies, using attributes in elements, other elements to "watch". Note that by "element", I mean any
      HTML tag. For example <input type="text" id="myText" ...> is an "element" -- into which one would insert various attributes.

      You can have many "elements" watching many other elements. In fact, elements can watch elements that are also watching other elements!

      The whole shebang is driven by changes to a "watched" element. Right now, changes are first detected in HTML "form" elements: <input>, <textarea>, <option>,
      and <select> (that's a limitation imposed by browsers -- in fact, if that limitation didn't exist this "watch" libarary wouldn't be that useful).
      So, when one of these changes (say, a button is clicked or text is modified),  a change occurs.
      We call the element that changes the "changedElement".

      When such a change occurs, anything that is "watching" such a changedElement will be "processed" -- as described below, one or more functions
      will be called.  These elements that are "processed" are called "thisElement". Actually, many elements can be watching a particular
      "changedElement", and each will be processed in turn. Thus, the "thisElement" refers to the element currently being processed.

      Note that an element can "watch itself". That means when the element changes, it will be processed (for example, one could
      change an entered number's decimal precision).



C.   Initialization

        initWatches MUST be called (say, in an onLoad function).

          * initWatches(doInit,doAlerts,nmax)

             MUST be called for "watch" to work!
             It scans all elements in the document for "watch", etc.
             It also adds appropriate onChange (or onClick) elements  to elements that are user changeable (such as <input> elements)
             and that are "watched" (technical note: that use doWatchedBy as the event handler)

             doInit: if  1 (the default), an "initialization" is done: process all elements that are watching anything (that have a watch attribute)

                To suppress this, use initWatches(0). Note that during processing, the "changedElement" will be 0 (as noted above,
                this can be a useful trick for suppressing processing during initialization).

             doAlerts: if 1, <alert> tags will be converted (this is done before watch elements are processed). Default is 0

             nmax: maximum number of elements to process. Default is 28000. This is meant to check for infinite recursion

D.  Implementation

      Uses attributes in the elements that are "watching for changes".  The key attributes are "watch" and "watchCall"

      Syntax:
          watch="id1 id2 ..."
          watchCall="func1 func2 func3(arg1,arg2,..) .... "

      Example:

          Let's call the element containing a "watch" and "watchCall" attribute "element X".

              The idN (id1, id2, etc) are ids of elements to watch (or, as described below, they could be "flags" or "names").

              If any of these "watched" elements change, the funcN functions in watchCall are called (in order) -- see below for how these functions are called

              Note: if element id2 is an <input type="text" ..>, and its value is changed:
                        X would be the "thisElement"
                        id2 would be the "changedElement"

              Special values of idN:
                      1   = watch self.  When the element's value changes, the watchCall functions are called.
                             Note: 1 is ALWAYS treated as the first element to check (regarcless of where in the list it is).
                                   So, on aesthetic grounds, we recommend specifying it first.

              If watchCall="" or watchCall is NOT specified -- and ndec not specified (see below) --  watchCall="useValue" is used.
              This will copy the value of the matching element to this element.

D.i.    Where to put these attributes?
           watch and watchcall attributes can be placed in any element! This includes input elements, textarea, and regular display (such as <span> and
           <div>).

           Keep in mind that a cascade of  changes  occur must start with a change to a user-changeable element: such as an <input> or <textarea>.
           However, once started the chain of changes (a watches b watches c ...) can incorporate any elements (such as <span>s and <div>s)
           For more details, see the Notes section below.


D.ii.      Specifying a value to watch for
           By default, any change to a "watched" element will lead to processing (calling the watchCall functions) of any & all  element that are "watching".
           In fact, sometimes (like for <input type="button"..>) just clicking will do this (even if no value is changed).

           If you wish, this processing can occur only when the changed element achieves a certain value (or one of a specified set of values).

           To do this, in the watch list use   idn=v1,v2,...
           The v1, v2, etc can be any string composed of numbers, characters, and some simple other characters. You can NOT include &,(, ),*, commas, or " inside
           of a v1,v2..  And, there can NOT be ANY spaces: not in the "v1", or next to the commas.

           Example:  watch="myEyes  herEyes=1  bothEyes=1,All "   watchCall="fixIt "

           In this example: any change to "myEyes" will cause the "fixit" function to be called, or
                             if "herEyes" is changed to 1, "fixit" is called, or
                             if "bothEyes" is changed to either 1 or All, fixit is called.

           Note that the comparision is case-sensitive, but spaces will be stripped.

D.iii.       Priority

            When an element changes, the elements that "watch it" are processed. The order they are processed is somewhat arbitrary; though "thisElement"
            is done first -- if an element is watching itself (has a "1" in the watch list), it will be processed first.

            Sometimes order of processing matters. For example:
               If B watches A, and set's it's own value to A + random number
               C watches A and B, and set's its value to A + B.

               In this case, when A changes B should be procssesd first, and then C (otherwise C might be computed using the value of B before it was updated)

            You can dictate order of processing by including a priority="n" attribute.
            Elements with LOWER values of n will be processed first: they will be earlier in the list of "watchedBy" elements
            associated with an element that changes.

            Note:
              * If no priority is specified, the element is given a priority of 1000  (i.e.; not listing a priority implies a priority of 1000)
              * Thus, a priority > 1000 is processed after elements with no proiority
              * thisElement (the element that changed) is always processed first, no matter what the priority value is.
              * if two elements are watching the same element, and have the same priority, the order of processing is indeterminate.
              * it is up to you to keep track of priority (to make sure that an element that should be processed earlier has a
                lower priority than an element that should be processed later)


              Example (assume that addRandom is a function that adds a random number to thisElement)
                  Enter value: <input type="text" id="A" ndec="2">
                              ...
                  Your value + random number: < span id="B" watch="A" watchCall="addRandom " priority="1"></span>
                        ...
                  Sum:  < span id="C" watch="A B" watchCall="sumWatch " priority="2"></span>

               This ensures that when A is changed: first, it is checked for validity; then B is processed; and then C.

D.iv.     Syntax of watchCall functions

         The watchCall functions should be custom functions you specify, or you can one of the built in functions.

         They will be called with 6 arguments:

             thisElement,changedElement,nthCall,arglist,orig_id,cElem

             thisElement: the element that is "watching" (i.e.; "X" in the above example) -- it is the element whose value will be changed

             changedElement: the element that was changed -- that is being watched. Note that "thisElement" may be watching several elements

             nthCall : the count, in the list of functions, of this function. .1 is used for the first in the list
                      Note:  the global watchData['call'][thisElement] is the list of calls

             arglist: array of arguments specified in the watchCall (will be empty if no arguments specified in the watchCall -- see below for details)

             orig_id : element that "started" this string of changes.
                      One important use of this is useful in detecting initialization (orig_id=0 during initialization)

             cElem : if changedElement was an element, will be the same as checkElement
                     if changedElement was a flag, cElem will be an object with fields value, tagname, and flag
                       value: time stamp (or the =n value)
                       tagName : always will be "_flagVar";
                       flag: name of the flag (same as changedElement)

                       Note: get_elemValue will return the value of changedElement in both cases (an element or a flag).


            When a "self" change is being processed:
                   thisElement and changedElement will be the same

             Notes:
                 * changedElement may not be an ID
                  In particular, on initialization (using initWatches),  the changedElem sent to watchCall functions will be "0".

                 * You can specify an "arglist" for any of the functions in a watchCall list.These will be sent to the custom function.
                    The arglist should be a comma delimited list of strings, surrounded by ().
                    For example:
                       watchCall="myFunc1(1,3)  myFunc2(abc,51, cat)  youFunc "
                    Note you should NOT use quotes around the arguments. Everything is treated as a string
                    The argument list will be parsed at the commas.

                        Thus, in the above example: the arglist sent to  myFunc will have a 2 elements, myFunc2 will have 3 elements, and youFunc will have zero elements.

                    Note: to include odd characters in the arguments, use html encoding.
                       &quot; for "  ,  &lt; for < ,  &gt; for >  , &amp; for &
                   and some non-standard ones:
                      &comma; for ,  ,   &lparen; for (, &rparan; for ) , &apost; for '

                      Example: myFunc( 12, &lparen; x &rparen; ,  3 &comma; a )  -- would yield argument 1 =12, argument 2=(x), and argument 3 = 3,a

D.v.     Returns from watchCall functions

           The simplest case is to return the new value of the element.  The element's value (on screen) will be updated, as will its old value.

            Or, you can exercise greater control by returning an object: with fields "value" and "status", and (possibly) "message"
            Example:
                ares={}; ares.value=newValue; ares.status='save' ;
                return ares;
            or, more simply
               return {status:'save',value:newValue} ;

              value should be a string, or a number (not an object).

              message: display (using alert) the message if  status contains "error" or "warning"

              status can have one or more of the following values (in a space delimited list). They are done in order of appearance

                   save : save newValue to the element, and to oldValue. Continue processing. You MUST specify a value (if you don't, undefined is used)

                   ok: save value of element to oldvalue. Continue processing. You should NOT specify a value with ok.

                   orig:  revert to original value -- the value before the first function was called (i.e.; the entered value)
                          This also reverts the "old value" to the prior old value. It implies "no change"

                   reset : set the value to its old value (value field is ignored). Of course, don't update old value
                          It implies "no change"

                   nosave: don't save anything (retains current value, does NOT update oldvalue ). However, any changes made by
                           earlier functions are retained.


                   skip: stop processing this element: don't do any more functions.  If this is encountered, no other status options are considered
                        This can be useful if an <input> element's value can either be set by the user or computed from the values of other elements.
                        To allow for this "computation from other values", the <input> element would "watch" these other  elements.
                        If you want to ensure that a value explicitily entered by the user is used, rather then a computed value, you
                        should "skip" the "computation from the values of other elements".
                        A convenient way to check for this  "change to self" is to  compare (in the function that does the computation from other elements)
                        thisElement and changedElement -- if they are the same, this element was directly changed (so "skip" processing).

                   jump:  jump the next 1 (or more) function calls. This is used by jumpIfSame

                   quit : stop processing this element, and do NOT process any other elements that may be watching the "changedElement".
                          Eg; this is used by checkNuumber to stop processing if a bad value was entered (and to reset to prior value)

                   warning: display the message

                   error: display the message, and then abort javascript (without any more changes)
                         This is a clumsy abort: javascript will report an "no such function: abortJavaScript" error

              Notes:

                 * status="", and not specifying status, are synonyms for status="save"

                 * you only MUST specify one of save, ok, orig,reset, nosave, quit, or error
                   quit by itself implies noSave -- however, you can specify quit with something else (such as save)
                    Order of precedence (if you specify more than one (from highest precedence to lowest)
                       error, save, ok, reset, orig, nosave

D.v.1.      Simple example of custom function:

                  function   doit1(thisElement,changedElement,nthCall,arglist,orig_id) {
                      var aval=get_elemValue(changedElement);
                      aval=aval/100 ;
                      return aval ;

           Slightly more complex

                  function   valueK(thisElement,changedElement,,arglist,orig_id) { .... }
                      var aval=get_elemValue(changedElement);
                      newval=aval/100 ;
                      if (aval<100) {
                          return {status:'reset'}   // illegal value, so revert to prior value. Hence, no need to specify a value field
                      }
                      return  {value:newval,status:'save'} ;
                   }

E. Features

    Several shortcutts and features are supported using special attributes. These include flag, update, alert, confirm, and number entry specifications

E.i: Alerts and confirms

      The alert and confirm attribute are alternatives to using alert(message) or confirm(message) in a watchCall list.
      In particular, if the message is long, it might be prettier to specify the alert (or confirm) in a seperate attribute,
      rather than embedded in a watchCall attribute.

      Furthermore, when specified in an attribute, alerts and confirms will  ONLY be done if thisElement changes.
      That is, if someother watched element changes, alert and confirm attributes are  ignore.
      In contrast, alert and confirm specified in a watchCall list will happen regardless of what was changed.


     Confirm:
       Given that the element changes and is being processed, confirm will display the message and give an ok or cancel
       (or something like that) choice.

          ok:  The value of the lement is retained, and the next function is called
          cancel: The value of the element reverts, and no more functions are called
                  The value reverted to depends on whether or not a default attribute was specified
                      default specified:   the default value is used
                      not specified: the value of the element before the change (the "old" value) is used

      Example: <input type="text" value="10" default="6" confirm="Are you sure you do not want to use the default value of this.default "
                watch="1 altInput" watchCall="modifyMe " >

        A) the user enters a value in the text box.
        B) The message (Are you sure .... valle of 6) is displayed in a confirmation box
        C) If the user accepts (eg; clicks on OK), then the current value of the element is retained,
           and the next function in the watchCall list is called (the modifyMe function)
           If the user cancels this, then the value is changed to 6, and no more functions are called


      Notes:
         * If both alert and confirm are specified, confirm is called first (and then alert)
         * When specified using attributes, alerts and confirms are processed before any explicitily specified function.
           In particular: after a checkNumber (specified using a ndec attribute) , and before  an update (specified using an update attribute).

         * As with function arguments (see D.iv above) you can include the &xxx; "entities" in the message (such as &comma; for  ','  &lparen; for '('
         * For details on how to include variable values in the message, see the alert and confirm function descriptions (below)
         * Note that if a confirm (or alert) are specified in the watchCall list, a confirm (or alert) attribute is ignored.
         * Alerts and confirms, when specified using attributes, are never processed during initialization
         * If confirm is used as an attribute, and a default attribute is specified, and the element's value equals this default, no confirm
           is shown (the entry is automatically accepted)

E.ii: Number checking with ndec

       The checkNumber function will check an element's value to make sure it is a legal number. If not, it will reset
       (to the "old value"). checkNumber will also transform the value, such as by truncating digits after a decimal.
       checkNumber can also check for bounds -- preventing entries outside of a min -- max range.

       To specify options for checkNumber, you must specify several attributes. In other words, checkNumber does not use arguments.

       Of particular interest is the nDec attribute. Ndec specifies the level of decimal truncation; thuse ndec="2" means a value entered
       as 100.2 will be displayed in the input box as 100.20.

       When ndec is specified, and checkNumber is NOT specified in the watchCall list, checkNumber will be added to the beginning
       of the watchCall list -- even if no watchCall list was specified!

       Thus, if you just want to check that an entry is a valid number, you can use ndec="xx", and not bother with a watch or a watchClass
       attribute.

       More precisely:

             If you specify an ndec attribute, then two things happen:

               If checkNumber is NOT in the watchCall list,
                     a) checkNumber is added to the beginning of the watchCall list
                     b) "1" is added to the beginning of the watch list

             IOW:  if you just want to check that a valid number was entered:
                   you can use ndec="whatever" (along with min, max, and any of the above attributes) and NOT bother with "watch" or "watchCall" elements.

             Notes:
                *  if you include checkNumber anywere in the watch list, then checkNumber will be done in sequence
                   (a "do checkNumber first" instruction is NOT added).
                *  You can include checkNumber several times, in the call list (say, if you want to format a value returned by an earlier function).
                *  Reiterating: if an ndec attribute is speified, then checkNumber will be called (even if no watch or watchCall attribute is specified)
                *  ndec="*" is special: it means "process checkNumber, but don't change the number of digits displayed". This can be useful
                   if you just want to check min, max, and other attributes

       For more details, see the description of checkNumber below.

E.iii    EVAL : evaluating simple expressions
          Sometimes a simple transformation is needed. For example, to express a weight in tons, you would divide by 2000
          You can create custom scripts, but for simple transformations the EVAL function can be more convenient.

          EVAL must be used with a eval attribute.  Unlike other "special" attributes (such as confirm), the eval attribute
          only is used to hold information. In particular, the expression to be evaluated.

          Thus, the placement of the EVAL function influences when the expression (stored in the EVAL attribute) will be stored.

          For now, you can only store one expression: so every time you EVAL appears in the watchCall list, the same expression is
          calculated.  One could put if statements in the eval, but if that is required it is probably easier to write a custom function.

          Since the eval function is in an attribute, there is no need to use the special entities (like &lparen;). However, the
          variable names supported by alert, confirm, etc are supported:  this.value, this.default, this.id, changed.value, and changed.name.

          Reminder: there MUST be spaces before and after these variable names.
             Note that the beginning and end of the string are considered to be spaces.
             Thus: "5+this.value" WILL not work ("5 + this.value" will work).

          Any legal javascript expression can be in the eval. You can included several lines of javascript code, seperated by semi-colons.

          The value used will be the last line evaluated in the EVAL.
          Thus eval="a=5-3 ; b=6*2 "
          or eval="a= this.value  * 6 ; a/10 "  -- if the element's value is 4, this returns 1.0

          Note that the results of eval are returned as is. That meays you can return "status" information, such as:
          eval="{status:'reset'} "

          If EVAL is in the function list but there is no EVAL attribute, an error occurs.


E.iv:   Flags

          In addition to "watch" lists, you can specify a list of "flags". These flags are variables that are updated after "thisElement"
          has been procssed (after  the functions in its watchCall list have been called)

          Thus, they are something of a substitue for "watch" -- rather then some 2nd element looking at a changed element, the changed element notifies
          other elements it has changed by upgadting a "flag" variable.

          This can be convenient if an element has to watch a lot of other elements. Rather than specyfing a long watch list, just watch a flag (use the
          exact same syntax as watching an element) -- and then watch the flag.  Any number of elements can update this flag!

          To specify a flag, use a flag attribute followed by the name of the flag(s) to updated (in a space delmited list).
          Examples:
                   flag="priceModifiers"
                   flag="pSet1  jset3 "

         When a flag is changed, an element that is watching it will be processed (just like if a <span> element it was watching had changed)
         In functions called The "changedElement" argument will be an object with three fields:
            tagName="_flagVar" (always)
            flagName=flagName  (name of the flag)
            value=timeStamp  (the second-stamp when the change occurred).

          You can directly check these fields, or you can use get_elemValue (whicl will the time stamp "value").

          Notes:
                 * flag updating occurs after all the element's functions have been processed
                 * a flag may be updated even if no values were changed. Thus, setting a flag means "we considered this element, even if we did nothing"
                 * flags can NOT have a value that is used as an id or name anywhere in the document (an error will occur).

          One-time initialization
             If you want explicitily set a flag's value, you can specify it using flagname=aval. Aval is simple: it can NOT contain any spaces or other
             odd characters. And there should NOT be any spaces before or after the = sign.

             Whenever this element is processed the flag value will be examined.
                 If it is already equal to "aval", the flag is not touched.
                 Otherwise, its value is set to aval (rather then the current time stamp).

             This can be a useful to initialize "just once" (say, you want to show an alert the first time something happens, but NOT on later occurence).

             Example:   flag="specificChange  generalChange=1 "

                 Whenever the element with this flag is "processed", generalChange is set to 1 (so after the first time, it doesn't change), and
                 specific change is set to the current timestamp (millisecond accuracy)


E.v:   updates

      Typically, an element (thisElement) will watch a few other elements, and when one of these watched elements changes, thisElement will be modified
      (using the functions thisElement's watchCall list).  You might have a number of such "thisElement and watched element" sets, but there isn't a lot of overlap.

      However, there might be situations where there are two sets, potentially large, of elements.

          a) A set of elements that may be changed directly. Call them the INPUTS
          b) A set of elements that will be updated whenever ANY element in the "INPUT" set is changed. Call them INFO elements.

          In this case, one could set up a an elaborate set of "watch" and "flag", but it could become cumbersome. Instead, using "Update" might be easier.

          Updating is accomplished by including an update="id1 ..." attribute in a number of elements. Just like a "watch", whenever an
          element with and Id or Name in this id list (or a flag name in this id list) changes, functions are called.
          In particular: when an update attribute is used -- the "checkUpdates" function is called.

          Note: just like elements in a watch list, you can specify values to look for by using "idn=avalue" syntax.

          checkUpdates will return a value stored in a global database, using thisElement's id as an index into the database.

              Basically, update="id1 ..." is a shortcut for: watch="id1 ..."  watchCall="checkUpdates".  In fact,

                 *  if you specify checkUpdates anywhere in  your watchCall list, the value of the update attribute will be added to the watch list, and
                     checkUpdates will be called like a normal function (in  other words, after any function appearing before it)

                 *  if you specify an update, and specify a watchCall that does not contain checkUpdates, checkUpdates will be added
                    to the beginning of the watchCall list (actually, an implicit checkNumber due to an ndec attribute will be placed before a checkUpdates)

         But how is data entered into this global database?  Using a custom built "doUpdates" function!

         You must write a doUpdates function, and it should save data to the watchData['updates'] object. The fields of this object should be
         the ids (not name or flag!) of elements that contain an update attribute (or contain a checkUpdate in its watchCall list).

        Actually, doUpdates should use the following functions to work with these watchData['updates'] values:
             update_elemValue(aelem,aval,aval2)  -- updates an element's value, and the 'updates' value, using aval. If aval2 specified, update the 'updates' with aval2
                                                    This can be useful if you want to display a truncated version (say 1.25) of a longer value (say 1.25614612)
             saveUpdateWatch(ado,aval,igroup)    -- set the ado 'updates' value = to aval. If igroup specified, save a "group specific" version of ado
             readUpdateWatch(ado,adef,igroup)   -- read the ado 'updates' value. If igroup specified, read a "group specific" version of ado. If no such 'updates' variable, return adef

         This doUpdates function will, typically, update the value of several (perhaps much more than several) values when called. That is: doUpdates
         should update all of the information displayed by the INFO elements.   doUpdates, just like any watchCall function, will be given
         the thisElement and changedElement arguments. However, typically a doUpdate will look at a number of values (say, of the INPUT
         elements) when computing updated values (say, for use in the INFO elements).  See the get_elemValue function (described below)
         for the recommended method for reading values from elements).

             Even more actually, the doUpdates function can be as flexible as need be, it can choose which value to update, based on what changes.

         But how is this operationalized?
             Reviewing, there are two steps.
                First: include update attributes (or, checkUpdate in a watchClass list) that lists what ids/flags/names to watch
                Second: a doUpdates function is created that will update values stored in watchData['updates']. This function can
                use any information to change any number of values that will be used by these elements with update attributes.

             The third step is executing this doUpdates function.  This can be done by creating an element (say, a SPAN) that watches
             a flag set by all the INPUT elements.  This element should also set a flag -- in particular, a flag included in a
             update list (i.e.; if  the flag1 flaag is set, and some other element includes an update="flag1" attribute).

             Thus, the logic is:
                 a) A number of input elements are created that set the same flag.  For example:

                    <input type="text" value="" title="First value" id="value1" flag="valueSetFlag">
                    <input type="text" value="" title="2nd value" id="value2" flag="valueSetFlag">
                    <input type="text" value="" title="3rd value" id="value3" flag="valueSetFlag">

               b) An element that will run doUpdates is created. For example

                  <span style="display:none" id="doUpdates_value" watch="valueSetFlag" watchCall="doUpdates " flag="updateDoneFlag" ></span>

                   doUpdates will change variables under doWatches['updates'] -- preferably using the functions list above. In particular, it
                   will change variables corresponding to INFO elements

               c) And a number of INFO elements that watches (using an update attribute) the "run doUpdates" element. For example:

                     Sum of values: <span id="sumValues" update="updateDoneFlag" ndec="2">
                     Product of values: <span id="prodValues" update="updateDoneFlag" watchCall="format(comma,dec0)" >
                     Difference: <span id="diffValues" update="updateDoneFlag" >
                     Minimum: <span id="minVal" update="updateDoneFlag" >

               Basically, the element that calls doUpdate is a central point -- it watches for any changes in the "INPUT" elements, and
               then sets in motion events that inform the "INFO" elements that their values have been updated.

       An example of a doUpdates.

             function doUpdates(thisElement,changedElement,icall,alist,origId) {
                var v1,v2,v3;
                v1=get_elemValueNumber('value1',0);
                v2=get_elemValueNumber('value2',0);
                v3=get_elemValueNumber('value3',0);
                watchData['updates']['sumValues']=v1+v2+v3;
                watchData['updates']['prodValues']=v1*v2*v3;
                watchData['updates']['diffValues']=v1-v2-v3;
                watchData.updates.minVal=Math.min(v1,v2,v3);
                return {status:"nosave"}
              }


        Notes:
            * This example will lead to all 4 of the INFO elements being updated whenever  any of the 3 INPUT elements change

            * The update flag used in the INFO elements is NOT the flag set by the INPUT elements. Instead, it is the flag of
              the <span> that calls doUpdates.  This is important: it guarantees that the updating will occur AFTER doUpdate has finished!

             * thisElement and changedElement can be used to determine which, of a set, of value to update.
               In particular, one could have several "run doUpdates" elements, each of which watches a different flag.
               The thisElement can be use to dictate what should be done.
               Note that in the above example, thisElement would have a value of "doUpdates_value".

             * Be sure to save values to watchData['updates']['xxx'] (or, watchData.updates.xxx), where xxx is case-sensitively the same
               as the id of the element to be updated.

             * Note the use of the return {status:"nosave"}. If you wish, you could change the value of this"dall doUpdate <span>:
               (say, display a count of the number of times its been called) in the usual fashion .

             *  Note the use of style="display:none" in the example (to hide this span no matter what is returned).

             * Assuming you don't want to display a value when doUpdates is called, you can put this "run doUpdates" element anywhere. For example
               if you have 3 distinct sets of INPUT,INFO elements, you could create three of <spans> (that watch the appropriate flag) at the
               end of your html file.

             * If you want to specify a default value to use if doUpdates has not created a value for an element,
               specify checkUpdates(defValue) in the watch list. If you use the update attribute and don't
               place a checkUpdtes in the watch list, the default value will be ''.


E.vi:   Saving parameters to, and restoring from, a file

      You can save the current parameters to a file, and then reload them at a later date.
      Due to javascript security, this is a bit roundabout --
        a) To save current parameters, you must manually save the contents of a window/tab to a file
        b) To load a saved set of parametes, you must cut and paste the contents of such a file to a <textarea>

     Two functions are provided, that can be called (say, using onClick of a button). Both are called with no arguments (or with a dummy argument)
        doSaveSpecs   -- save parameters to a new window, which the user can then manually save to a file
        doRestoreSpecs   -- set paramter values, given a "saved parameters file" manually pasted to a textarea

     doSaveSpecs will look for ANY tag with a 'save="xx"' attribute, and save it's ID and value. Note that the value is "escaped", so new lines and other odd charcters will
     be retained.
       Thus, you MUST indicate which parameters should be saved by including this save attribute!

     doRestoreSpecs will restore the parameters identified in such a saved paramteres file.

     How the restoration is done depends on the value of the save parameter.  The following are supported:
          VALUE  (or V) : restore the value of the element
          CALL   (or C) : if a watchCall is defined for this element, call the functions. The id of the element is used as the thisELement and changedElement, and origID will be "1"
          WATCH  (or W) : look at all elements this is watching and check for  a change -- if change detected, process this element (using the changed element as changedElement)
          BY     (or B) : process all element that are watching this element (that this element is 'watchedBy' )
          DELAY ( or D) :  process this element after all other elements have been processed (even it appears earlier in the HTML file).
          You can combine the above in a space delmited list (the order doesn't matter).  For example: save="V C B" would simulate a normal change of a "watched" elements value.

          Notes:
                * a VALUE by itself can limit redundant processing (say, it there are several elements  that call the same function)
                * a CALL by itself  is a useful way to simulate a button click (since no value needs to be changed).
                * a WATCH is useful if a watching element is large and shouldn't have its value touched, but might have visiblity effected (say, using a showIf).
                  Note: WATCH uses just the first element in the WATCH list that is not "self" as the "changedElement".
                  Note that you CALL can be used to process the element, or BY (which will process "self")
                * BY is useful if this is a "signal" element (i.e.; one used in an UPDATE)
                * WATCH and BY are effected by order of appearance: earlier elements are processed first (hence a WATCH may be processed before an element in the WATCH list is changed)
                  Use DELAY for a workaround ...
                * DELAY:  If more than one element has a DELAY, process them in order of appearahce


     Notes:
       * each entry in a saved parameters file has the format:
            saveValue=idName= valueToUse
        The saveValue is the "xx" in a save="xx", the idName is matched to an element, and the valueToUse is used.

      * lines beginning with // are treated as comments, and are ignored

                    reset
F. Special tags

   Several special tags are supported: <alert>, <more>, and <help>

F.i:   <alert>

     If  you specify 1 in the 2nd argument of initWatches, <alert> elements will be processed.

     These should have  attributes:
        id : the id
        on: id(s) and/or flags to watch. If it changes, the alert is displayed (turned on)
        off: id(s) and/or flags  to watch. If it changes, the alert is hidden (turned off)
        init: if init="show" is specified, this alert will be always be shown  right after initialization  (it will start "on")
              if init="no" is specified, this alert will NOT be shown right after initialization (it will start "off")
        zap="message"  : if specified, a button will (with value message) precededthe alert -- when clicked, it hides just that alert
        alertButton="aid" : id of the open-box-of-alerts button (see below)
        class="aclass" : class used in <div> that contains this alert. If not specified, "alertLine" is used.

    Within the <alert> ... and </alert> you can place your message using standard html

    Specifying a "open-box-of-alerts" button.
        A special attribute is used to create a button, or link, that will open the set of alerts:
           use="aid"
        aid should point to an element (i.e.; a <div>) that contains the <alert>s.
        When a use attribute is specified, the following elements are relevant:

          class and style: class is always used when displaying the button. style can be overridded by onStyle (as described below)

          id:   The id of  this "view the alerts"  element  should be used in alertButton attributes of alerts -- this signals that the
               button (or link) should be highlighted when this alert is turned on.

          onStyle:  Contains the css-style to use when the alertButton is highlighted.
                    If onStyle is not specified, a "onStyle="border:4px solid red" is used.

           value: If specified, a button will be created using this as the text.
                  If not specified, a link will be created using   the contents between <alert use=".." ...> and </alert>

                  Sorry, FireFox limitations mean you MUST have an </alert> after <alert>, even if there is nothing in between!

           init="1"  : if init="1" is specified, than the button (or link) will be "highlighted" on initialization (the onStyle is used as the style)

           title, class, style: used in the surrounding span. Note that style is used when "button" is NOT highlighted. Class is always used.

        Note that when the "open-box-of-alerts" button (or link) is clicked, the highlighting will go away (whatever style you specified will be restored).

        Technical note: the button or link that actually opens the "set of alerts" is embedded in a span. The span is what is highlighted.

F.ii:   <help>

     <help> creates a button (or link) that toggles display of help text in a special "moveable" box.

    Example:
         <help title="the full gas price details" style="color:blue" value="??"   use="fullGasPriceHelp"> </help>

     Sadly, <help ....  /> doesn't seem to work in firefox, so you MUST follow <help ...> with and end tag of </help>

     Attributes used in <help>

        Value:   If there is nothing between the <help> and the </help>:
                 a value attribute is read and used as the text of a <input type="button" ..>

                 If no value attribute is specified:
                     If there is nothing between the <help> and a </help> :  "?" is used
                     If there IS somethingbetween the <help> and a </help> :  a  linke (<a >..</a> is used (with the stuff between <help> and </help> the
                     link's text)

     use:  id of a container whose contents should be displayed. For example, the id of a <div>.
     style: is optional: if specified it is used. If <a> method is used, a default style of "text-decoration:underline;color:blue"
          is used (if no style attribute is specified)
    title: is the mouseOver title. If not specified, "Help" is used




F.iii:   <more>
    Creates button or link that when clicked, toggles display of a DIV/SPAN element.

    Example:   <more title="the full gas price details" style="color:blue" value="??"   use="fullGasPriceHelp"> </more>

   Sadly, <more ....  /> doesn't seem to work in firefox, so you MUST follow <more ...> with and end tag of </more>

     Attributes used in <more>

  Value:
     If there is nothing between the <more> and the </more>:
          a value attribute is read and used as the text of a <input type="button" ..>
          If no value attribute, "..." is used
    otherwise
        a <a > is used (with the stuff between <more> and </more> the text)
  Attributes used by <more>
    use: the ID of the element (Div, span, whatever) to toggle display of.
    altUse : instead of toggling display, simulate a click of another <more> -- a <more> with an id equal to altUse's value (note: altFor is a synonym for altUse)
    value: if specified, text displayed in a button (see above for details)
    altValue: if specified, the "value" displayed (in button or in link) is alternated between the value and this value.
    style: optional: if specified it is used. If <a> method is used, a default style of "text-decoration:underline;color:blue" is used (if no style attribute is specified)
    title: the mouseOver title. If not specified, "Help" is used
    id: per usual.  This is used when an altUse attribute is specifed (in a different <more>

  Note: one of altUse or use should be specified, not both. If both are specified, altUse is used.
        altUse is a convenience: it allows you to have multiple closers, where one of the closers is the "main" closer. This "main" closer
        can modify itself using "altValue" and "altBorder".
        For example, if you have <more> buttons that close the same box, the message of the "main" close button could get out of sync.
        Using altUse avoids this problem.

  Example:
         <more value="button text"  use="sheet1" altValue="Close gas worksheet" title="Open the cost-per-gallon-of-gas worksheet">link text</a>


G.    useful functions

    * doWatch(cElem,nmax,noIter)

         Find all elements watching cElem, and process them. In other words, act like cElem was "changed".
         This allows you to simulate a "watched" change on elements that you (via javascript) directly modify --
        including flags, and <span> and other such "non-user-changeable" elements.

         cElem: the id (a string) of the "changed element". This can refer to a flag, or an ID or NAME of an html element
                If nothing watches this flag or element, nothing is done.
         nmax : stop procesing after this many elements. Default is 20000. This can be used to check for infniite loops (see note below)
         noIter : if 1, do NOT recurse. Default is 0 (recursion IS done).

         Recursion means:
             Say an element (call it X1" changes.
             If an element (call it Y1) is watching X1, first process Y1 (call all the functions in Y1's watchCall list),
             and then also process all elements that watch Y1".
         Recursion is very powerful: it is how changes are propagated!
         However, recursion can lead to infinite loops (say if X1 watches X2 and X2 watches X1 and they both
         change whenever processed). That's why nMax can be useful

         Examples:
          The html:
             <input id="check1" type="checkbox" title="click me for 100" watch="1" watchCall="doinit1" >
             <span id="info10" title="summary value here">...</span>
             ...
             <span id="info10w" title="repeat of info10" watch="info10" >...</span>
             ..
         A custom watchCall function
            function doit1(thisElement,changedElement)
              set_elemValue('info10','100')
              doWatch('info10')
              return {status:'nosave'}
            }

         In the above,
           a) check1 is clicked
           b) doit1 is called
           c) doit1 change the value of info10 (using set_elemValue)
           d) by calling doWatch (with a info10 as the argument), a "watched change" of info10 is simultated -- all elements that watch info10 will be processed
           e) that means that info10w will be processed. In  this case, since there is no watchCall, the value of info10w will be set to info10's value (i.e.;
              the default watchCall of "useValue" is used.


   * avalue=get_elemValue(anid,adefault)

      Read the value of the anid element (the element with an id of anid). This can be any element, it does NOT have to be an element mentioned in a "watch" list.
      If there is no such element, return adefault.

      get_elemValue is sort of smart.
         If anid is a flag, will return the flag's value (typically, the time stamp when the flag was last changed)
         anid can be a pointer to a node (an object), or a string containing the id of an element. Or, if no such flag or id exists, a name.
         The "name" is only used for "radio" buttons.

        For <input> elements
            text and textarea: returns the value
            checkbox: if checked, returns the value. If not checked, returns the value of a "noCheck" attribute. If there is no noCheck attribute, returns adefault
            radio: returns the value of the checked radio button (that has the same name as the radio button with an id of anid). If no radio button (in this set is checked),
               returns the value of a "noCheck" attribute (in the anid element), or adefault.
               Note that this means you MUST have an id in one (but not necessarily all) of the set of radio buttons you wish to check (sorry, refernence by name not yet supported)
            option or select: returns array of selected values
         For other elements,
            returns the innerHTML

   * avalue=get_elemValueNumber(anid,adefault,toInt,ndec,blankOk,stripdollar)
     avalue= get_elemNumber(anid,adefault,opts)

     Both are similar to get_elemValue, but will return a number type (not a string).

     If the value of anid is not a number (note that'' is considered to be not a number), adefault and blankOk are used.
       If the trimmed value = '', and blankOk=1, return ''
       If adefault is nan, return adefault
       Otherwise, use adefault

     If anid's value is a number, or adefault is used (where adefault is a valid number):

        If ndec is specified, the number of decimal point digits is set (i.e. value.toFixed(ndec) is called -- ndec<0 means "round to nearest power of 10")
        If toInt =1, the value is converted to an integer (uing parseInt). Otherwise, a floating point number is returned (using parseFloat)

       Note that ndec conversion is done first. This guarantees that a number type, and not a string type, is returned.
       Yes, ndec=0 and toInt=1 are kind of redundant, so be it.

     The opts in get_elemNumber  correspond to the following in get_ElemValueNumber

         get_elemNumber       get_elemValueNumber
         blankOk                 blankOk=1          (the default is 0 : '' is treated as nan)
         int                     toInt=1            (the default is 0 : a floating point number is returned
         dec=n                   ndec=n             (the default is '' : no decimal point truncation
          NOT$                   stripdollar=1       (do NOT strip $ from from front -- if not specified, learing $ is removed)

      Note: internal commas are always removed.

   * anAtt=readAttribute(anid,attlist,adef,notrim)
     anAtt=readAttributeNode(anode,attlist,adef,notrim)   (use pointer to node (an object), not value of id (a string)
           Read the value of an attribute from an element.

           and: id of an element to look at
           attlist:  a list of attribute names to look for. The first matching attribute is  used (case insenstive search).
                    Thus, you can use attlist to specify a bunch of "alternate names" for this attribute
           adef : default value, if this attribute does not exist
           notrim: if 1, don't trim. Otherwise, trim spaces before returning value

   * updateFlag(aflag,flagValue)
          Set the aflag value to be flagValue. If flagValue not specified, current time is used

 * set_elemValue(aid,avalue)
     Set the value of the aid element, to be avalue.
     set_elemValue is sort of smart. If aid is a checkbox or radio, then will "check" it if the element's value equals avalue. That is, it will NOT change the value of the element - it will check it.

G.i   Built in functions that can be used in watchCall

  *  alert(message)                      Display message in an alert box.
  *  avgWatch and avgWatch0              Compute an average of all elements thisElement is watching.
  *  background(oColor,sColor,altId)     Change the background color when something changes.
  *  checkNumber                         Check that thisElement's value is a legitimate number (used by ndec, etc)
  *  confirm(message)                    Display message in a confirmation box.
  *  display(id1,id2,...)                Hide or display a list of elements.
  *  fade(aid,totalSecs,secStep)         Fade (or fade in) an element
  *  format(opt1,opt2,...)               Similar to checkValue, but for use with purely displayed values
  *  hideIf and showIf                   Hide or display thisElement (given value of changedElement)
  *  jumpIfSelf(ido,njump)               Skip the  next njump functions if thisElement==changedElement (or !=).
  *  noDisplayIf(aname,aref,cond)        Conditionally hide elements with name="aname"
  *  prodWatch and prodWatch0            Similar to sumWatch, but product (x1*x2...)
  *  skipIf(arg)                         Skip the remainder of  this watchCall list under certain conditions. arg can be init or notinit
  *  sumWatch and sumWatch0              Sum of all the elements in thisElement's "watch". sumWatch0 skips missings and NaNs.
  *  useValue                            Returns value of the changedElement (i.e.; to be used  thisElement's value)

G.i.a   Built in functions that can be used in watchCall:details

    * alert(message)

       Display message in an alert box. Message can include commas.
       In addition, there are 5 special strings: case sensitive and must be surrounded by spaces (beginning and end of string count as a space)
          this.id   : the id of thisElement
          this.value  : value of the element (i.e.; the user entry
          this.default : default attribute (if one was specified). If no default attribute, '' is used
          changed.id : id of changedElement (possibly a flag name)
          changed.value  : value of changed dlement

       Example: alert(the this.id current value is  this.value, and the changed element, changed.id, was changed.value)

       Notes:
          * this is a fairly simple function, so don't put parenthesis, quotes,   or other non alphanumeric characters in the message

          * alert is most useful as a debugging tool!

          * Instead of using alert(message) in a watchCall list, you can specify an alertattribute. As with checkNumber,
            the confirm will be called early (after checkNumber and before update), using the message.

          *  If alertis specified in the watchCall is, an alert attribute is ignored.


   * avgWatch and avgWatch0

       Compute an average of all elements thisElement is watching.
           avgWatch: bad values (blanks or non-numbers) cause a warning message and no change.
          avgWatch0 bad values are ignored
       See sumWatch for further details


   *  background(otherChangeColor,selfChangeColor,selfAltId)

       Change the background color when something changes.
       If the thisElement changes, or if the  "changedElement" equals selfAltId, the background color is set to selfChangeColor
          selfAltId can contain * -- it will be replaced by the element's "group" attribute's value.
       If some watched element, other then the above, changes: the background color is set to otherChangeColor

       The colors can be names (such as "red", or "yellow"), or hex value (such as "#efefef" or "#ab31cc")

       Thus, this can be used to alert the user (by setting the background to otherChangecolor) that something has changed, so you should modify an element...
       and then remove this "alert" after the element (or its alternate) is modified  by setting the background color to selfChangeColor

    * checkNumber : check that thisElement's value is a legitimate number

       If thisElement's value is not a number, then an error message is displayed and the original value is displayed.

       Hint: to format "non-input" elements, consider using the "format" function described below

       checkNumber can do several other things, such as formatting the display and checking against a minimum.
       These other actions are specified by including attributes in thisElement

       The supported attributes are as follows. Note that if the value is not a number, the attributes are not checked:
       a "not a number" error message is displayed, and the prior value is restored:

          ndec="n"
                is the number of decimal points to display (after entry)
                if n="*", don't modify decimal digits, but check everything else (such as min and max)
          min="amin"
          max="amax"
                 Check if value greater-than-or-equal-to amin and/or less-than-or-equal-to amax. Displays error message if one (or both) fail, and reverts to prior value.
          prec="pvalue"
                 Precision. Round to this precision. For example: prec="10" will rounded to this level of precision (so 133 is displayed as 130)
          default="value"
                Default value. Used if a '' is entered.
                If not specified, '' is treated as a non-number (hence will be an error)
          blankOk="1"  : if blankOk="1" is specified, than empty cells do NOT cause an error.

          Shortcut:
               You can embed commas (eg; 1,315.12), or use "K " and "M" (case insensitive) notation eg: 1.35k or 1.613M).
               The values will be converted to numbers (that is, commas and "k" and "m" are removed).
               Hence, the above examples would become 1315.12, 1350, and 1613000

          Hint: for an alternative way of calling checkNumber, see section E.ii above.

    * confirm(message)

       Display message in a confirmation box.  As with alert, message can include commas and 5 "special strings" (this.id, etc)

       If the user clicks on ok (or yes), confirm returns status:'ok' (so the next call is done)
       If the user clicns on no (or cancel), confirm returns status:'reset' (revert to values BEFORE element was changed)

       Instead of using confirm(message) in a watchCall list, you can specify a confirm attribute. As with checkNumber,
       the confirm will be called early (after checkNumber and before update), using the message.
       Note that if confirm is specified in the watchCall list, a confirm attribute is ignored.

       See above for the details.

    * display(id1,id2,...)

        Will hide or display a list of elements.
        The ids (id1, etc) should be element names, with a "-" before elements that should be hidden, or a @ sign before elements
        that should be toggled (between displayed and hidden).
           Or, if an no element with exists, names will be searched -- so all elements with this name will be displayed (or not displayed).

        Shortcut: "1" (or -1 or @1) refers to "this element"

        If no ids are specified, nothing is done.

        Example:
           <input type="checkbox" watch="1" watchCall="display(row1,-row2,-row3,@helpRows)" title="Just see row 1"   />
           would display the row1 element, hide the row2 element, and toggle display of the helpRows element (between visible and non-visible)

        display  will determine the type of the element, and use the appropriate display attribute (i.e.; "block" for <div>, "inline" for span,
        and "table-row" for <tr>).

        Hint: if a "group" attribute is specified, then the idList will be modified.
        In particular, any occurence of a '*' will be replaced by the value of the group attribute.

        Example:
           <input type="checkbox" group="3" watch="1" watchCall="display(rowA*,-rowB*,-rowC,@help)" title="Just see row 1 of this group"  />
        will be treated at:
           <input type="checkbox" group="3" watch="1" watchCall="display(rowA3,-rowB3,-rowC,@help)" title="Just see row 1 of this group"  />

       Example:
           <input type="checkbox"  watch="1" watchCall="display(-helpName,help1)" title="Hide all help except the first"  />

    * fade(id1,totalSeconds,nSecs)
    
        Fadeout (or fade in) an element
        
        id1 : The id of the element to fade in or out.
               Special values:  "."  (without the quotes):  fade this element
                                ".."  (without the quotes):  fade the parent of this element
        totalSeconds : time span of the fade, in seconds. I.e.; "30" means "a 30 second fade"
                       If 0, or less than 0, "stop" an active fade of id1 (and leave in current opacity)
        nsecs:     : time between fade, in seconds (fractional seconds okay). I.e. "5" means "change opacity every 5 seconds"
                     If nsecs<0, this is a fade in -- start invisible, and fade in to full visibility.
                     Nsecs>0 0 is a fade out -- start fully visible, and fade out. When 0 is hit, element is hidden (display='none')

        If fade is called on an element undergoing a fade, the prior fade is cancelled and a new one begins.

       Hint: some useful combinations
          totalSeconds = 0  : just cancel prior fade (and retain current opacity)
          totalSeconds=0.1 and nsecs=0.1  :  fade out  quickly (basically,  equivalent to a style.display="none")
          totalSeconds=0.1 and nsecs=-0.1  : fade in quickly  (basically, make visible)


    * format(opt1,opt2,...)      (or formatNumber)

        Similar to checkNumber, but for use with purely displayed values (that will not be checked by checkValue). That is, this is typically
        used in non <input> elements!
        The following options are supported:
           trim   : remove leading and trailing spaces
           upper  : convert to upper case
           lower  : convert to lower case
           comma  : add commas
           K      : express as X.xxK   If <100, use nnn
           M      : express as X.xxM   If <1000, use nnn. If <1,000,000 and > 1000, use n.xxK
           decN   :  round to Nth  decimal digit (i.e.; dec3 is similar to ndec="3")
           INT    : as integer. This overrides decN. Note that dec0 and int are the same thing
           cash or $   : express as $nnn.mm
           % or pct: convert decimal to pct. Ie. 0.43 becomes 43%

           if both k and m are specified: m is used if value > 1,000,000. Otherwise, k is used.
           if k or m used: everything but comma is ignored (comma applied to the X portion)
           if cash used, everything (except comma) is ignored

           Examples:
              format(comma,dec3)
              format(cash) (or format($) )
              format(M)

       Note: trim will work with non-numbers.

   *  hideIf and showIf
       Hide this element. In contrast to display and noDisplayIf, this only effects thisElement (the element containing this call)

       Note: see showif for the "display" complement to hideIf

       At least one, and possibly more, arguments are given.
       The current value of the changedElement --- NOT the value of thisElement -- will be compared to these arguments
       (case-sensitive comparision with spaces trimmed).
       If one of them matches, the element will be hidden (or, if showIf) displayed.

       Example:
         <span id="a1"  watch="whichModel" watchCall="hideIf(1,2) showif(4,5)">This is model 4 or 5</span>
         <span id="a1"  watch="whichModel" watchCall="hideIf(4,5) showif(1,2)">This is model 1 or 2</span>

         Where whichModel could be a set of radio buttons with values from 1 to 5.
         Note that in the above example, a changedElement value of "3" will cause NO changes.

       Notes:
           * if changedElement=thisElement, nothing is done.

   * jumpIfSelf(ido,njump)

       Skip the  next njump functions if thisElement==changedElement.
       Default is to jump over (not execute) the next funtion (after jumpIfSelf in the watchCall list).
         ido:  1 : skip if same
               0 : skip if not the same
         njump: default is 1. # of function calls to skip. Must be integer >0.  If jump to the end, same as "skip"

          remainder of  this watchCall list under certain conditions. Basically, this function returns either status:"skip" or a status:"nosave"

        You can have several jumpIfSelf in a watchCall list. You can NOT jump backwards!


   * noDisplayIf(condList)

          Will conditionally hide elements. This is similar to display, since it will hide any element (as specified by their ids)

          The condList can have 3 elements (per usual, comma seperated):
              aname:  consider all elements with a name field equal to "aname" (aname can contain *, for use with a group attribute)
              aRef :  extract the "aRef" attribute from the elements being considered. If not specified, or if '', "refValue" is used
              cond: a condition to use. The value of aRef is compared to the value of thisElement (the element that contains the noDisplayIf).
                    If the condition succeeds, the elements is hidden.
                    The supported conditions are: lt le gt ge eq ne
                    If cond is not specified, gt is used.

          Note: thisElement value and refValue should be numbers!

              Example:
                      Assume a "3" is entered in an input element.
                      This input element has   noDisplayIf(myrows,refValue,gt) (or, noDisplayIf(myrows) )
                      And the following is specified:
                        <span name="myrows" refValue="1">This is 1 </span> |
                        <span name="myrows" refValue="22">This is twentyTwo </span> |
                        <span name="myrows" refValue="-31">This is -3 </span> |
                        <span name="myrows" refValue="2">This is -41 </span>
                     Then the "This is  twentyTwo" will NOT be displayed (since the value of refValue, 22, is greater than 3)
                     Note that elements with the name can be anywhere in the document, and there can be any number of them.

                     Importantly, as one changes the input element, the choice of what is displayed will change. So, if 0 were entered, then ONLY "This is -3" would
                         be displayed.

    * prodWatch and prodWatch0 : similar to sumWatch, but product (x1*x2...)

    * sumWatch and sumWatch0: the sum of all the elements in thisElement's "watch" list is returned (excluding "self")
                Bad values (blanks or non-numbers) cause a warning message and no change.
                If no watch list for the element, no change.

               sumWatch0 is the same as sumWatch, but bad values (blanks or non-numbers) are ignored.


   *  showIf
       Show this element. In contrast to display and noDisplayIf, this only effects thisElement (the element containing this call)

       The syntax is the same as in hideIf


   * skipIf(arg)
       Skip the remainder of  this watchCall list under certain conditions. Basically, this function returns either status:"skip" or a status:"nosave"

       arg can be:
           init :  if this is called during initialization, skip. Otherwise proceed (return a nosave)
           notInit : if NOT initialization, skip.

       The default value of arg is "init". Arg is case insensitive.


   * useValue : the value of the changedElement is used as the thisElement's value. Pointless on a self call!
     If changedElement is "0", do nothing (eg; as may occur during initialization)
     Note: this is the default if no watchCall (or if watchCall="") is specified


H.     Notes and reminders:

      * If any of the elements specified in a watch changes, the watchCall functions are done.
           The element (actually, the ID of the element) that changed is used (in calls to functions) as the changedElement.

     * An element with a "watch" can be watched!
       In other words, watching is recursive: after an element's "watch" list is processed (say, due to a change in an element it is watching);
       any element that watches this will also be processed.

        Example:
           Assume that
               * elementB watches elementA and elementA2
               * elementC watches elementB
            If,
              *  elementA  changes
            then ...
              *  elementB  will be processed  (elementA will be the changedElement)
              *  elementC  will be processed (elementB will be the changedElement)

            And this "recursion" happens automatically!


       For example:
         Var1: <input type="text" id="elementA"> and Var2: <input type="text" id="elementB">
         Sum of var1 and var2 =  <span id="elementAB" watch="elementA elementB" watchCall="sumWatch ">
           ( == <span id="elementC" watch="elementAB" watchCall="format(K) "> )

      *  As illustrated in the above example, non-input elements (that do not have a native onchange capability) can be "watched".
         Changes to them, due to an "input" element they are watching, will be treated properly.
         However, direct changes (say, using your own javascript code), will NOT be detected.

 */

var watchData={};
watchData['oldValue']={};
watchData['watch']={};
watchData['watchedBy']={};
watchData['call']={};
watchData['flags']={};
watchData['flagsChanged']={};
watchData['updates']={};
watchData['temp']='';    // might be useful
watchData['timeouts']={};    // might be useful


//==================
// intisilze watchData etr
function initWatches(noInit,doAlerts,nmax) {

var doAlerts,alist,funcAlts,list1,im,aelem,awatch,addCheckNumber,anDec,aFlag,aid,aname,igroup,awatchCall,tmp1,doself,wargs,tmpz,doselfargs,lookees={},toWatch={} ;
var iavv,gotCheckNumber,ffuncs,tmp,ffargs,itmp,atmp,rrlist,tatmp,ff,aid1,blist,blistn,jm,wid,ipriority,clist,jm2,aa ,avv2,avv,aUpdate,gotCheckUpdates,addCheckUpdates;
 var addAlert,addConfirm,aAlert,aConfirm,gotAddAlert,gotAddConfirm ;

if (arguments.length<2) doAlerts=0;
if (arguments.length<3) nmax=28000 ;

// ref :: <div  class="alertLine"  watch="fullGasPrice" watchCall="skipIf(init)  display(-alertGasPrice)" id="alertGasPrice"></div>

if (trim(doAlerts)=='1') {               // convert <alert>...</alert> -- basically, replace with a div with a watch and watchcall attributes (that is then processed below)
  alist=document.getElementsByTagName('alert');
  convertAlerts(alist);
}                                                           // doalerts

var funcAlts={DISPLAY:"doDisplayElement",FORMAT:"formatNumber",NODISPLAYIF:"noDisplayIf",HIDEIF:"hideIf",JUMP:'jumpIfSelf',NONE:'doNothing',FADE:'fadeAnElement',
              SHOWIF:"showIf",BACKGROUND:'setBackgroundColor' ,SKIPIF:'skipIfInit',ALERT:"showAnAlert",CONFIRM:"showAnConfirm",EVAL:"doWatchEval"};

list1=document.getElementsByTagName('*');

//  first, extract flags and check for * oddness
initProcessFlags(list1) ;
//goo='';
//  now process each element that contains a watch attribute  :::::::::::::::::::
for (im=0;im<list1.length;im++) {

   aelem=list1[im];

// get watch list an watchCall list
   awatch=readAttributeNode(aelem,'WATCH','');  awatch0=awatch;

   aFlag=readAttributeNode(aelem,'FLAG','');
   if (trim(aFlag=='')) aFlag='0';

   addCheckNumber=0;
   anDec=readAttributeNode(aelem,'NDEC DEC ','');
   if (anDec!=='') addCheckNumber=1;

   addCheckUpdates=0;
   aUpdate=trim(readAttributeNode(aelem,'UPDATE UPDATES CHKUPDATE CHKUPDATES ',''));
   if (trim(aUpdate)!=='') {
       addCheckUpdates=1 ;
       awatch=aUpdate+' '+awatch;
   }

   addAlert=0;
   aAlert=trim(readAttributeNode(aelem,'ALERT ',''));
   if (trim(aAlert)!=='')    addAlert=1 ;

   addConfirm=0;
   aConfirm=trim(readAttributeNode(aelem,'CONFIRM ',''));
   if (trim(aConfirm)!=='')    addConfirm=1 ;


   if (addCheckNumber==1 || addAlert==1 || addConfirm==1 ||  aFlag!=='0') awatch=' 1 '+awatch;

   if (awatch==''  ) continue ;    // not a watched element

   aid=trim(aelem.id) ;
   aname=readAttributeNode(aelem,'name','');

  igroup=readAttributeNode(aelem,'group','');
  if (trim(igroup)=='') {
      if (aid.indexOf('*')>-1 || aname.indexOf('*')>-1)  abortJavaScript('Error: in '+aid+' you included a * in either the ID or NAME, but you did NOT specify a group attribute');
  }

   if (trim(aid)=='')    abortJavaScript('Error: an element has a WATCH list, but does not have an ID (watch list= '+awatch+'\n'+aelem.outerHTML);

   watchData['watch'][aid]=[] ;


// process the watch list   (that might of been augmented due to alert, update, etc attributes
   tmp1=getWord(awatch);
   doself=0;
   for (jm=0;jm<tmp1.length;jm++) {
           avv=tmp1[jm];
           wargs=[];
           if (avv.indexOf('=')>-1) {
               tmpz=avv.split('=');     // for now, ignore any =xx stuff
               avv=tmpz[0];
               wargs=tmpz[1].split(',');
               for (var itmp=0;itmp<wargs.length;itmp++) wargs[itmp]=trim(wargs[itmp]);
           }
           if (avv=='1') {               // if watching self, always check self first (so changes to self trump changes elsewhere)
                 doself=1;
                 doselfargs=wargs;
           } else {
                 avv2={ID:avv,ARGS:wargs};
                 iavv=document.getElementById(avv);
                 if (iavv) {                                    // id exist
                     lookees[avv]=iavv;
                     watchData['watch'][aid].push(avv2) ;
                 } else {
                    iavv=document.getElementsByName(avv);
                    if (iavv.length>0) {
                       lookees[avv]=iavv[0];
                       watchData['watch'][aid].push(avv2) ;
                    } else {                              // no such id, perhaps a flag (do NOT put an entry in lookees list)?
                       if (typeof(watchData['flags'][avv])=='undefined'){
                          alert("Error in WATCH: no such flag or ID (  "+avv+') in watch list (or update list) of '+aid);
                       } else {
                           watchData['watch'][aid].push(avv2) ;      // check, but do NOT add event, to a "flag"
                       }
                    }
                 }
            }
   }                               // stuff in watch list

   if (doself==1) {                         // add "self" element if "1" in the watch list
        avv2={ID:aid,ARGS:doselfargs};
        watchData['watch'][aid].unshift(avv2);   // updates of self, specified with a 1, are always checked first
        lookees[aid]=1;
   }

//  create call list for this aid.

// get watchCall list
  if (readAttributeNode(aelem,'WATCHCLASS CLASSWATCH','')!=='')  abortJavaScript('Warning:  for '+aid+' you specified a watchClass attribute (you probably meant watchCAll ');
   awatchCall=readAttributeNode(aelem,'CALL WATCHCALL CALLWATCH CALLS WATHCALLS CALLWATCHS CALLWATCHES ','');

   watchData['call'][aid]=[];

   gotCheckNumber=0;
   gotCheckUpdates=0;
   gotAddAlert=0; gotAddConfirm=0;
// parse tmp for function names and argument lists. Argument lists are surrouned by () (after function name), with , seperateing arguments.
// you can use html entties, such as &comma; and &lparen; and &lt; to insert problemmatic characters (the will be converted to the actual character)
  tmp=initProcessWatchCall(awatchCall,funcAlts);
  ffuncs=tmp[0]; ffargs=tmp[1];

// move these functions and arg lists (and check)
   for (var itmp=0;itmp<ffuncs.length;itmp++) {
         atmp=ffuncs[itmp];
         rrlist=ffargs[itmp];

         tatmp=trim(atmp.toUpperCase());   // a shortcut function name?
         if (typeof(funcAlts[tatmp])!=='undefined') atmp=funcAlts[tatmp];

         if (!window[atmp])   abortJavaScript('Error in watchCall for id='+aid+': no such function: '+atmp);

         if (atmp=='checkNumber') gotCheckNumber=1;
         if (atmp=='checkUpdates') gotCheckUpdates=1;
         if (tatmp=='ALERT') gotAddAlert=1;
         if (tatmp=='CONFIRM') gotAddConfirm=1;

         ff=[atmp,rrlist];
//         if (aid=='createColl_3') alert(' adding to call: '+ff);
         watchData['call'][aid].push(ff)
   }                                    // dont with call procsssing

// add calls to start of watchCall list (due to special attributes). Note that the of calling is reversed (later entries are called eatlier)
// But all of these are called before any explicitily listed function
   if (addCheckUpdates==1 && gotCheckUpdates==0)    watchData['call'][aid].unshift(['checkUpdates']) ;     // update attribute set and no checkUpdates function call
   if (addAlert==1 && gotAddAlert==0)  watchData['call'][aid].unshift(['showAnAlert1']) ;
   if (addConfirm==1 && gotAddConfirm==0)  watchData['call'][aid].unshift(['showAnConfirm1']) ;
   if (addCheckNumber==1 && gotCheckNumber==0)   watchData['call'][aid].unshift(['checkNumber']) ;  // start with checkNumber if ndec specified and checkNumber NOT in watchCall list

// updateFlag is called AFTER all other functions
   if (aFlag!=='0') {
     watchData['call'][aid].push(['updateFlag']);       // add flag processor function (as  epecified by flag="xx" atribute -- do it AFTER all other functions
//      if (aid=='createColl_3') alert(' flag adding to call  ');

//     goo=goo+'\n updateflag adde dfor '+aid;
   }
   // no call at all.. insert updateValue
   if ( watchData['call'][aid].length==0)  {
//      if (aid=='createColl_3') alert(' usevaule adding to call ');
     watchData['call'][aid].push(['useValue'])
   }
// add to watchedBy lists .....

    for (var jm=0;jm<watchData['watch'][aid].length;jm++) {
            avv2=watchData['watch'][aid][jm];
            avv=avv2['ID']; warg=avv2['ARGS'] ;
            if (typeof(watchData['watchedBy'][avv])=='undefined') watchData['watchedBy'][avv]=[];
            avv2={ID:aid,ARGS:warg};
            watchData['watchedBy'][avv].push(avv2);                                            // element avv is being watched by  element aid
            if (typeof(lookees[avv])!=='undefined') toWatch[avv]=lookees[avv];     // towatch is used below (lookees used to ignore flags
    }

}                               // all elements with a watch  ::::::::::::
//alert(goo);

 sortWatchedBy(1) ;

// add onchange or onclick to all "watched " elements that are input, textarea, or select (deal with names too!)

for (aa in toWatch) {           // aa is an id of an element that is being watched
  addEventsToWatch(aa) ;
}


if (trim(noInit)!=='1') {
  tmp=doWatchMany('0','*',nmax,0)  ;      // initialize
  if (tmp>nmax)   alert('Warning: too many elements processed ('+nmax+'). There may be an infinite loop');
}
//alert('asfasf '+ watchData['call']['createColl_3'].join(' | '));
return 1;

}

//=======================
// add change or click event to a "watched" element (might be several elements if Name, not id, is refered to
function addEventsToWatch(aa) {
   var daid,daids,jj,atag,ovalue;
   daid=document.getElementById(aa);      // id trumps name
   if (!daid) {
      daids=document.getElementsByName(aa) ;
   } else {
      daids=[daid];
   }
   for (jj=0;jj<daids.length;jj++) {
      daid=daids[jj];
      atag=daid.tagName.toUpperCase();
      if (atag=='INPUT' || atag=='TEXTAREA' || atag=='SELECT') {
         atype=readAttributeNode(daid,'type','').toUpperCase();
         if (atype=='BUTTON' || atype=='RADIO' ) {                     // note: click on a selected radio will be treated as a change event
            daid.onclick=doWatchedBy ;   // add onchange only to user changeable elements
         } else {
            daid.onchange=doWatchedBy ;   // add onchange only to user changeable elements
         }
      }
      ovalue=get_elemValue(daid,'');            // get_elemValue will deal with "name" properly
      watchData['oldValue'][aa]=ovalue;
   }                                            // daids
 return 1;
}

//=============================
// sort watchedBy using priority
function sortWatchedBy(i) {
var aid1,alist,blist,blistn,jm,wid,ipriority,clist,jm,jm2,wtmp,wargs,jm3,im,ords,pgot ;
 for (aid1 in watchData['watchedBy']) {
       alist=watchData['watchedBy'][aid1];   // a watchedBy list -- elements that watch aid .... who should have first crack?
       blist=[];
       blistn=[];
       clist=[];
       pgot={};
       for (jm=0;jm<alist.length;jm++) {
          wtmp=alist[jm];                    // element of a watcher
          wid=wtmp.ID ;
          wargs=wtmp.ARGS ;
          if (wid==aid1) {             // self: always first
            blist[0]=[wtmp];
            pgot[0]=1;
            continue ;
          }
          ipriority=readAttribute(wid,'PRIORITY ORDER',1000);
          if (trim(ipriority)=='') {  // no priority always done last   DEPRECATED: use priority of 1000
              blistn.push(wtmp);
              continue;
          }

          if (isNaN(ipriority))  abortJavaScript('Error: priority value for '+wid+' is not a number (was '+ipriority);
          if (ipriority<1) abortJavaScript('Error: priority value for '+wid+' must be greater than 0 (was '+ipriority);

          ipriority=parseInt(ipriority);
          pgot[ipriority]=1;

          if (typeof(blist[ipriority])=='undefined') blist[ipriority]=[];
          blist[ipriority].push(wtmp);
       }                                             // elemets in aid's watchedBy list
       ords=[];
       for (im in pgot) ords.push(im);

       if (ords.length>0) {
          ords.sort(function(a,b){return a-b}) ;
          for (jm=0;jm<ords.length;jm++) {
            jm2=ords[jm];
            for (jm3=0;jm3<blist[jm2].length;jm3++) {
              wtmp=blist[jm2][jm3];
              clist.push(wtmp) ;
            }
         }
       }
       for (jm2=0;jm2<blistn.length;jm2++) {
          clist.push(blistn[jm2]) ;
       }
       watchData['watchedBy'][aid1]=clist ;    // save the "sorted by prioirty" list
 }                                          // aid
 return 1;
}

 //================
// process a watchCall list
function initProcessWatchCall(tmp) {
  var ffunds,ffargs,dds,boo,dds,aa0,mm2;

  ffuncs=[];
  ffargs=[];

  while(1==1) {
    if (tmp.indexOf('(')<0) {
        if (trim(tmp)=='') break;
         dds=getWord(tmp);
         for (mm=0;mm<dds.length;mm++)  {
            ffuncs.push(dds[mm]); ffargs.push([]);
         }
         break;
    }                            // else, found a (
    boo=parseUsing(tmp,'(',')');
    dds=getWord(boo[0]);
    for (mm=0;mm<dds.length;mm++)  {
      ffuncs.push(dds[mm]); ffargs.push([]);
    }
    aa0=boo[1].split(',');
    for (mm2=0;mm2<aa0.length;mm2++) aa0[mm2]=someHTMLDecode(aa0[mm2]);
    ffargs[ffuncs.length-1]=aa0;
    tmp=boo[2];                      // back to top and process this remainder
  }
  return [ffuncs,ffargs] ;
}


//==================
// proces flags
function initProcessFlags(list1) {

  var im,aelem,aFlag,agroup,myFlags,ixm,bflag,tmp   ;
  for (im=0;im<list1.length;im++) {
   aelem=list1[im];
   aFlag=trim(readAttributeNode(aelem,'FLAG FLAGS UPDATEFLAG ','0'));
   if (aFlag=='' || aFlag=='0') continue;

   agroup=readAttributeNode(aelem,'group','');
   if (agroup!=='') aFlag=aFlag.replace('*',agroup);

   myFlags=getWord(aFlag);
   for (var ixm=0;ixm<myFlags.length;ixm++) {
     bflag=myFlags[ixm];
     if (bflag.indexOf('=')>-1) {
          tmp=bflag.split('=');     // for now, ignore any =xx stuff
          bflag=tmp[0];
     }
     if (document.getElementById(bflag) || document.getElementsByName(bflag).length>0) {
        abortJavaScript("Error: element:"+aelem.id+" contains a flag value ("+bflag+") that is the same as an element id or name ") ;
     }
     watchData['flags'][bflag]=0 ;   // initialize value to be 0


   }                             // all the flags in this flag="..."
  }
  return 1;
}                               // done proessing flags


//-====================================
// convert <alert> into <div> with watch, etc attributes
function convertAlerts(alist) {

var ia,iid,aid,atitle,ause,isbutton,avalue,dalink,aspan ;
var atitle,astyle,onstyle,aclass,doinit,onlist,offlist,azap,oninit,newnode,abutton,t2,cnodes,pnode,i,iac,znode;

for (ia=alist.length-1;ia>=0;ia--) {
     iid=alist[ia];
     aid=iid.id
     if (trim(aid)=='') {
       atitle=readAttributeNode(aelem,'title','');
       abortJavaScript('Error: an <alert> does not have an ID '+'\n'+aelem.outerHTML);
     }

     ause=readAttributeNode(iid,'use','');    // <alerts> with use are treated special
     if (ause!=='') {                            // if use specified, this is the "view the alerts button" spec
        if (!document.getElementById(ause))  abortJavaScript("Error: <alert> use value, "+ause+", does not point to any element's id");

        isbutton=1;
        avalue=readAttributeNode(iid,'value','');
        if (avalue=='') {
            avalue=get_elemValue(iid,'','S');
            if (avalue!=='') {
                 isbutton=0;
            } else {
              avalue="View";
            }
        }
        if (isbutton==1) {
            dalink=document.createElement('input');
            dalink.setAttribute('type','button');
            dalink.value=avalue ; dalink.setAttribute('value',avalue);
        } else {
            dalink=document.createElement('a');
            dalink1=document.createTextNode(avalue);
            dalink.appendChild(dalink1);
        }


        aspan=document.createElement('span');
        aspan.setAttribute('use',ause);
        dalink.onclick=function() {
            var istat=turnDisplayOn(this.parentNode.getAttribute('use'),'toggle',aid);
            bbstyle=this.parentNode.getAttribute('offStyle');
            this.parentNode.setAttribute('style',bbstyle);
            return false} ;

        aspan.id=iid.id ;
        atitle=readAttributeNode(iid,'title','View alerts');    aspan.setAttribute('title',atitle);
        astyle=readAttributeNode(iid,'style','border:1px solid #dedede'); aspan.setAttribute('offStyle',astyle);
        onStyle=readAttributeNode(iid,'onStyle',astyle);         aspan.setAttribute('onStyle',onStyle);
        aclass=readAttributeNode(iid,'class','');
        if (isbutton) {
             dalink.setAttribute('class',aclass);
        } else {
             aspan.setAttribute('class',aclass);
        }
        doinit=readAttributeNode(iid,'init','0');
        if (doinit=='0') {
            aspan.setAttribute('style',astyle);
        } else {
            aspan.setAttribute('style',onStyle);
        }

        aspan.appendChild(dalink);
        iid.parentNode.insertBefore(aspan,iid);
        iid.parentNode.removeChild(iid);
        continue ;
     }                                  // end of "use" attribute

// if here, a standard <alert>

     onlist=readAttributeNode(iid,'on','');
     offlist=readAttributeNode(iid,'off','');
     oninit=readAttributeNode(iid,'init',''); oninit=trim(oninit.toUpperCase());
     azap=readAttributeNode(iid,'zap','');
     aclass=readAttributeNode(iid,'class','alertLine');
     abutton=readAttributeNode(iid,'alertButton','');

     newnode=document.createElement('div');
     newnode.id=aid ;
     newnode.style.display='block';



     newnode.setAttribute('onlist',onlist);
     newnode.setAttribute('offlist',offlist);
     newnode.setAttribute('watch',onlist+' '+offlist);     // watches all of them, and then decides what to do based on which one changes
     newnode.setAttribute('alertButton',abutton);
     t2=''+Math.random(); t2='ID_'+t2.substr(2);
     newnode.id=t2;


     t2='skipIf(INIT) ';                        // do NOT process alerts on initialization (use the INIT attribute (show or noshow) value to determine visibility
     if (oninit=='SHOW') {
        newnode.style.display='block';
     }
     if (oninit=='NO' || oninit=='NOSHOW') {
        newnode.style.display='none';
     }


     t2=t2+' doAnAlert ';                   //doAnAlert will read onlist and offlist-- sort of a display with context sensitive list
     newnode.setAttribute('watchCall',t2);

     newnode.class=aclass;     newnode.setAttribute('class',aclass);

     cnodes=iid.childNodes ;   // now move everything from <alert> to this new <div>
     pnode=iid.parentNode;
     for (i=cnodes.length-1;i>=0;i--) {
         iac=cnodes[i];
         newnode.insertBefore(iac,newnode.firstChild) ;
     }

   if (trim(azap)!=='') {
      znode=document.createElement('input');
      znode.setAttribute('type','button');
      znode.value=azap; znode.setAttribute('value',azap);
      znode.setAttribute('title','click to hide this alert');
      znode.setAttribute('zap','1');
      znode.onclick=function(){turnDisplayOn(this.parentNode,'0')} ;
      znode.style.fontSize="60%" ; znode.style.color="red";
      newnode.insertBefore(znode,newnode.firstChild) ;
   }

     pnode.insertBefore(newnode,iid);                 // and put this div in the tree, and remove the <alert>
     pnode.removeChild(iid);

  }                                                      // do the next alert (in alist)

  return 1 ;
}                                                           // doalerts


//======================
// called when an input, textarea, or select element is changed, and it is being something (its id appears in a watch list)
// this uses watchedBy to lookup what elements are "watching" this, and then  processes each of the elements (calls the functions specified in each of the element's  watch list)

function doWatchedBy(e) {

 var targ,myid,do1,doit,a1,im,istat,checkme={},irec,newby,ndid0,ndid,awwby2;

  if (!e) var e = window.event;   // from http://www.quirksmode.org/js/events_properties.html
  if (e.target) targ = e.target;
  else if (e.srcElement) targ = e.srcElement;
  if (targ.nodeType == 3)    targ = targ.parentNode; //safari bug

  myid=readAttributeNode(targ,'id','');
  if (myid=='') {                         // should never happen ...
        abortJavaScript(' Error: watchedBy element has no ID ');
  }

// check for "directly watched" and "watched by name" (no need to check watched by flag)
 awwby=[]; awwby2=[];
 if (typeof(watchData['watchedBy'][myid])!=='undefined') {      // not directly watched. Maybe it's  "name" is?
    awwby=watchData['watchedBy'][myid] ;
 }

 if (typeof(targ.name)!=='undefined') {
    aname=targ.name  ;
    if (typeof(watchData['watchedBy'][aname])!=='undefined') {
       awwby2=watchData['watchedBy'][aname] ;
    }
 }

 for (im=0;im<awwby2.length;im++) awwby.push(awwby2[im]);
 if (awwby.length==0) abortJavaScript('Error: doWatchedBy evoked by '+myid+', but no watchedBy element exists.');

 // each element that is watching this is "processed". Note that stuff on watchedBy is ALWAYS an id of an element (never a flag, never a name)

  if (document.getElementById('debugNow')) {
    if (document.getElementById('debugNow').checked) {
       document.getElementById('debugOutput').innerHTML= document.getElementById('debugOutput').innerHTML+'<br> in doWatchedBy, onchange event on:  '+myid ; // escape(document.getElementById(myid).outerHTML);
    }
  }

 ndid=doWatchMany(myid,awwby,2000);

 if (ndid>20000) {
         alert("Error  processing WATCH lists: too many (more than 20,000)! You might have an infinite watching loop ");
         return 0;
  }
  return 0;
}


//==============================
// recursively process a bunch of elements; myid is the "changed element" (on first pass)
// returns # of elements processed
// Recursion means: if one of the elements  (in awwby) changes, then process all elements it is watched by
// nmax is a counter (default value of 2000) -- if exceeded, return immediately. THe caller can check to see if "too many" elements were
// noIter is a flag (default value of 0). If 1, do NOT  do recursion (just process the awwBy list, don't worry about what may be watching elements in awwby
// procssed (which may be a sign of an infinite loop)


//    * doWatchMany(myid,alist,nmax,noIter)
//
//         Go through a list of elements, and process their watchCall functions. And recurse!
//
//         myid: used as the "changed element".  Default is "0" (i.e.; to signal initialization)
//         alist:  alist is a list of ids that contain a watchCall list. * means "do all elements with a watch, th (it will be the thisElement in calls on its watchCall list).
//           That is, for each entry in alist:
//                a) The entry, call it aid,  will be the thisElement in function calls
//                b) The watchCall list of aid is procesed
//                c) each function on this watchCall list is called,  with aid thisElement, and myid as changedElement
//         nmax : stop procesing after this many elements. Default is 2000. This can be used to check for infniite loops.
//         noIter : if 1, do NOT recurse. Default is 0
//         orig_myid : "original" myid.  Can be used in special cases (such as parameter restore)
//
//         Recursion means "if an element (call it Y1)  in the watch list is changed, then process all elements that watch Y1".
//         This is used to have changes propagate! However, it can lead to infinite loops (say if X1 watches X2 and X2 watches X1 and they both
//         change whenever processed). That's why nMax can be useful
//
//         Examples:
//                   doWatchMany()    -- used in initialization
//                   doAllWatches('VALUE1 VALUE2 ','VALUE0')   -- Element value1 and value2 will have their watchCall functions called, with VALUE0 as the "changed element"
//
//       Note: doWatchMany can be called in a timed loop using setTimeout -- so that elements are updated regularly.
//       That's kind of inefficient, hence we recommend careful use of watch instead.



function doWatchMany(myid,awwby,nmax,noIter,orig_myid) {

 if (arguments.length<5) orig_myid=myid ;                          // the original element that was changed (helps detect initializatio)
 if (arguments.length<4) noIter=0;
 if (arguments.length<3) nMax=2000;
 if (arguments.length<2) awwby='*';
 if (arguments.length<1) myid='0' ;

 var doit,im,a1,nrec,newby,im2,istat,athis,achange,blist,a,aaflag,stopit=0,istat0,vstat,stopit,skipit,athis2,wargs,tmp;
 var chkval,achange,iok;


 if (awwby=='*') {                                   // process all elements with a watch that watches itself
      blist=[];
      for (a in watchData['watch']) {            // all elements that have a watch
        tmp=watchData['watch'][a]  ;
        for (ja=0;ja<tmp.length;ja++) {           // all elements an element is watching
            atry=tmp[ja];
            if (atry.ID==a) {           // self watch only!
               blist.push(atry);
            }
        }
     }
     awwby=blist;
  }


 doit=[];
 for (im=0;im<awwby.length;im++) {
    a1=awwby[im];
    if (myid==0) {
        doit.push([a1,a1.ID]);
    } else {
        doit.push([a1,myid]);     // do this "thisElement,changedElement" pair
    }
 }
 nrec=0 ;                           // recursion count .. stop if it gets too big

 while (1>0) {                         // exit with a break
   newby=[];
   for (im=0;im<doit.length;im++) {
      nrec++;

      if (nrec>nmax) {
         return nrec ;
      }
      athis2=doit[im][0]; achange=doit[im][1];    // achange is myid -- the id of the element being watched (that has a watchedBy list)
      athis=athis2.ID;
      wargs=athis2.ARGS;

// check for = arguments
      if (achange!==0 && wargs.length>0) {
          iok=0;
          chkval=trim(get_elemValue(achange,''));
          for (im2=0;im2<wargs.length;im2++) {
            if (wargs[im2]==chkval) {
               iok=1; break;
            }
          }
          if (iok==0) continue ;     // no match, so don't do.
      }


  if (document.getElementById('debugNow')) {
    if (document.getElementById('debugNow').checked) {
       document.getElementById('debugOutput').innerHTML= document.getElementById('debugOutput').innerHTML+'<br> in doWatchMany: calling dowatch1 with '+athis+' and ' +achange+' (orig ='+orig_myid;
    }
  }

//if (document.getElementById('debugNow').checked) alert(' call dowatch1 with '+athis+','+achange);
      vstat=doWatch1(athis,achange,orig_myid)  ;           // this will call the functions, and deal with saving/skipping, etc.

        istat=vstat[0];
        skipit=vstat[1];
        stopit=vstat[2];

        if (stopit!='0' && orig_myid!==0) {
             return nrec ;      // a quit (no more watch processing-- Ignored on initialization
        }

        if (istat>0) {
          if (athis!==achange || orig_myid==0) {     // on init (orig_myid=0), process stuff watching athis. Non-init: if athis is the "self" in a longer list, no need to process again
             newby.push(athis) ;  // thie element changes, so might need to recurese   (process stuff watching athis)
          }
        }

                                                           // track number of elements (prevent infinite looops)
   }

   if (trim(noIter)=='1') return nrec ;                                 // no "recursion"?


 // ready to recurse!
 // 1)  check on elements that watch elements just changed (awwby list)
 // 2) check on elements that watch flags that may have been updated

   doit=[];                   // the list of elements to "recurse over" -- start building it

   for (im=0;im<newby.length;im++) {
       achange=newby[im];                             // an element that was processed, and changed, in the above
       if (typeof(watchData['watchedBy'][achange])!=='undefined') {
           for (im2=0;im2<watchData['watchedBy'][achange].length;im2++) {         // what elements are watchedBy this "changed" element
               athis2=watchData['watchedBy'][achange][im2]   ;
               athis=athis2.ID;  wargs=athis2.ARGS;
               if (athis!==achange)  {                // no need to reprocess elements that were procssed above
                   doit.push([athis2,achange]);     // do this "thisElement,changedElement" pair
               }
           }
       }
   }

   for (aaflag in watchData['flagsChanged']) {
       if (typeof(watchData['watchedBy'][aaflag])!=='undefined') {       // a flag that was set by elements processed above
           for (im2=0;im2<watchData['watchedBy'][aaflag].length;im2++) {         //   this "changed" flag is watchedBy what other elements
               athis2=watchData['watchedBy'][aaflag][im2]   ;
               athis=athis2.ID;  wargs=athis2.ARGS;
               if (athis!==aaflag)  {                // no need to reprocess elements that were procssed above (this should never happen)
                   doit.push([athis2,aaflag]);     // do this "thisElement,changedElement" pair
               }
           }
       }
   }
   watchData['flagsChanged']={};                                        // clear the list of flags changed

   if (doit.length==0) break ;                                             // nothing more to do!

 }                            // while loop

 return nrec ;

}


//===================
// process all the items on a watchCall for an element (if any exist).
// returns [nchanges,skip func calls, skip func calls and other watches]
// nchanges= 0 if this value was NOT changed, or how # times it was changed). Thus, value >0  means "thisElement was changed" (thought it might have been "changed" to the same value)

function doWatch1(thisElement,checkElement,orig_id) {
    var clist,ic,icc,vnew,newval,astatus,isSave,isNoSave,isReset,isQuit,acall,nchanges=0,oldval,isError,isWarn,amessage,isSkip,isOk,isJump  ;
     var celem,goo,clist1,arglist,ischange,isOrig,icdo;
   var startValue,startOldValue,goo;

    startOldValue=watchData['oldValue'][thisElement];   // used on a reset
    
   startValue=get_elemValue(thisElement,'');

// if (checkElement.substr(0,8)=='evSpecsR') alert(' dowatch1 this change '+thisElement+','+checkElement) ;

   if (typeof(watchData['call'][thisElement])=='undefined') return [0,0,0,0];
//     abortJavaScript('Error: no watch call list for '+thisElement);
// }


    clist=watchData['call'][thisElement];

    ic=-1;
    for (ic0=0;ic0<clist.length;ic0++) {
        ic=ic+1;                       // this is so jump will work cleanly
        if (ic>=clist.length) break ;   // beyond end
        clist1=clist[ic];
        acall=clist1[0];
        arglist=[];
        if (clist1.length>1)  arglist=clist1[1];

          if (typeof(watchData['flags'][checkElement])!=='undefined') {   // deal with flags that change (not elements)
               celem={};
               celem['value']=watchData['flags'][checkElement];
               celem['tagName']='_flagVar';
               celem['tag']='_flagVar';
               celem['flag']=checkElement;
               celem['flagName']=checkElement;
          } else {                                                  // this is the typical case
             celem=checkElement ;
          }

  if (document.getElementById('debugNow')) {
    if (document.getElementById('debugNow').checked) {
       document.getElementById('debugOutput').innerHTML= document.getElementById('debugOutput').innerHTML+'<br>'+ 'in doWatch1, call '+acall+'; thisElement '+thisElement+' and checkedElement ' +checkElement;
    }
  }


          vnew=window[acall](thisElement,checkElement,ic,arglist,orig_id,celem);                 // call the function (with argList, that might be empty)!

          if (typeof(vnew)=='object') {                                      // detailed return
              if  (typeof(vnew.status)=='undefined')  vnew.status='';
              if (trim(vnew.status)=='') vnew.status='SAVE';
              astatus='  '+vnew.status.toUpperCase()+' ';
              isSave=astatus.indexOf(' SAVE '); if (isSave<0) isSave=0;
              isNoSave=astatus.indexOf(' NOSAVE '); if (isNoSave<0) isNoSave=0;
              isOrig=astatus.indexOf(' ORIGINAL ');
                  if (isOrig<0)   isOrig=astatus.indexOf(' ORIG '); if (isOrig<0) isOrig=0;
              isReset=astatus.indexOf(' RESET '); if (isReset<0) isReset=0 ;
              isOk=astatus.indexOf(' OK ');
                  if (isOk<0)    isOk=astatus.indexOf(' OKAY '); if (isOk<0) isOk=0;
              isQuit=astatus.indexOf(' QUIT ');
                  if (isQuit<0) isQuit=astatus.indexOf(' STOP '); if (isQuit<0) isQuit=0;
              isSkip=astatus.indexOf(' SKIP ');   if (isSkip<0) isSkip=0;
              isError=astatus.indexOf(' ERROR '); if (isError<0) isError=0 ;
              isWarn=astatus.indexOf(' WARNING '); if (isWarn<0) isWarn=0 ;
              isJump=astatus.indexOf(' JUMP ') ; if (isJump<0) isJump=0;


// is jump is "internal", so do it first
              if (isJump>0) {
                   newval=vnew.value;
                   ic=ic+newval ;
                  continue ;
              }


              if (isQuit>0 && (isOk+isSave+isReset+isError)==0) isNoSave=1;                      // quit implies noSave (unless otherwise specified)

              if  ( (isSave+isNoSave+isReset+isQuit+isOk+isError+isSkip)==0 ) {     // must be a legit status
                    abortJavaScript('Error in element '+thisElement+' calling '+acall+ ' : status must include one of save, ok, noSave, reset, quit or error ['+astatus);
              }

              if (isError>0) {                            // fatal error!
                 amessage='Fatal error processing '+thisElement+' and '+checkElement;
                 if (typeof(vnew.message)!=='undefined') amessage=vnew.message;
                 alert(amessage);
                 abortJavaScript(1);
              }

              if  (isWarn>0) {
                 amessage='Warning while processing '+thisElement+' and '+checkElement;
                 if (typeof(vnew.message)!=='undefined' && typeof(vnew.message)=='string') amessage=vnew.message;
                 alert(amessage);
              }

              if (isSave>0 ) {
                 if  (typeof(vnew.value)=='undefined')  abortJavaScript('error: for '+thisElement+': saving result of '+acall+' , but no value field ');
                 newval=vnew.value;
              }

          } else {                             // simple return

              isSave=1; isQuit=0; isReset=0; isOk=0;
              newval=vnew;
          }

          if (typeof(thisElement['flagName'])=='undefined') {  // do NOT change a "flag variable" -- this should NEVER be a problem...
               if (isSave>0 || isOk>0 ) {                              // a "save this change": either "as entered" (ok) or using a newValue (save)
                    if (isOk==1) newval=get_elemValue(thisElement,'')  ;  // ok means "use the value as entered "(value BEFORE the function call)
                    watchData['oldValue'][thisElement]=newval;          // save means "use the value returned by function (might be the same as enterd value, might be different)
                    if (isSave>0 ) {
                        goo=update_elemValue(thisElement,newval);    // as returned by function
                        if (goo==0) {
                            abortJavaScript('Problem saving value of:'+ thisElement) ;
                        }
                        nchanges++ ;
                    }
               } else if (isReset>0) {                                       // reset goes all the way back to value BEFORE thisElement was changed
                    watchData['oldValue'][thisElement]=startOldValue;          //  reset goes back to value before change occurred
                    goo=update_elemValue(thisElement,startOldValue);
                    if (goo==0) {
                       abortJavaScript(thisElement.outerHTML);
                    }

                    nchanges=0;                                                // revert to original, hence it hasn't changed
               } else if (isOrig>0) {
                    watchData['oldValue'][thisElement]=startOldValue;          //  original goes back to values before first function called (i.e.; the entered value)
                    update_elemValue(thisElement,startValue);
                    nchanges=0;                                                // revert to original, hence it hasn't changed

               }  // do nothing if nosave
          }                                                   // flagname

          if (isQuit>0 || isSkip>0) break ;         // no more calls, perhaps no more watches
     }                                 // call ic
     return [nchanges,isSkip,isQuit,newval];
 }


// ================================
// useful functions (for use in watchCall list). Some of these are meant to be called using a synonym in the watchCall list (i.e.; format instead of formatNumber)
 //===================================

//=============================
// doWatch: a simplification of doWatchMany. Basically, this can be used by programmers (the above funtions are meant for internal use)
// myId : the element that "changed" ::   The "watchedBy" elements of myid are processed  -- where myId may be an ID, a NAME, or a FLAG
// nmax: max number of items to process (used to catch infinite loops)
// noIter: if 1, do NOT "recurse" (do not process elements that watch elements that watch myid)
// returns: space delimited string with list of elements that watched this

function doWatch(myid,nmax,noIter,orig_myid) {
var awwby,awwby2,aname,im,ndid,isflag,goo;
 if (arguments.length<4) orig_myid=myid ;
 if (arguments.length<3) noIter=0;
 if (arguments.length<2) nMax=2000;
 if (arguments.length<1) myid='' ;

 if (typeof(myid)!=='string' || trim(myid)=='')   abortJavaScript('Error in doWatch: myid is not a string:'+myid);


 awwby=[]; awwby2=[];  isflag=0;

 if (typeof(watchData['flags'][myid])!=='undefined') {           // flag of this name is watched?
    awwby=watchData['watchedBy'][myid] ;
    isflag=1;                                       // if a flag, don't bother looking for "name" or "id" watchedBy lists
 }

// check for "watching this flag", "directly watched" and "watched by name"
 if (isflag==0) {
     if (typeof(watchData['watchedBy'][myid])!=='undefined') {      //  directly watched.
        awwby=watchData['watchedBy'][myid] ;
    }

    targ=document.getElementById(myid);     // check if anything is watching the name of this element (in addition to the id of this element)
    if (targ) {
//      if (myid=='carType_1') alert (' goooo '+myid+' = '+targ+','+typeof(targ));
      if (typeof(targ.name)!=='undefined') {
         aname=targ.name  ;
         if (typeof(watchData['watchedBy'][aname])!=='undefined') {
            awwby2=watchData['watchedBy'][aname] ;
         }
         for (im=0;im<awwby2.length;im++) awwby.push(awwby2[im]);        // add to watching id list (so an element can have its ID and its NAME watched.
      }
    }
 }            // isflag

 if (awwby.length==0) return 1;                                     // nothing is watching this flag, or id or name

 // each element that is watching this is "processed". Note that stuff on watchedBy is ALWAYS an id of an element (never a flag, never a name)

 ndid=doWatchMany(myid,awwby,nmax,noIter,orig_myid);

 if (ndid>nmax) {
         alert("Error  processing WATCH lists: too many (more than "+nmax+" iterations)! You might have an infinite watching loop ");
         return 0;
  }

 goo='';
 for (im=0;im<awwby.length;im++) goo=goo+' '+awwby[im].ID ;

  return goo ;
}


//============================================================================
 // check if thisElement's value is a number. Note that this is called using the regular "watchCall" function syntax, but only thisElement (1st argument) is used.
 // So we only include one argument!

 function checkNumber(thisElement,changedElement) {
 var andec,amax,amin,aprec,adef,bok,myvalue0 ;
 var amess,myvalue,tt,tt1,tat1;
 var ovalue=watchData['oldValue'][thisElement];


 andec=readAttribute(thisElement,'ndec dec ','');

 if (trim(andec)=='*') andec='';                   // ignore it (andec='*' used as shortcut
 if (trim(andec)!=='') {
     if (isNaN(andec)) return {message:'Error: non-numeric value for ndec attribute in '+thisElement+': '+andec,status:"error"};
 }

 adef=readAttribute(thisElement,'default def','');
 if (adef!=='') {
     if (isNaN(adef)) return {message:'Error: non-numeric value for default attribute in '+thisElement+': '+adef,status:"error"};
  }

 amin=readAttribute(thisElement,'min minimum amin ','');
 if (amin!=='') {
        if (isNaN(amin)) return {message:'Error: non-numeric value for amin attribute in '+thisElement+': '+amin,status:"error"};
 }
 amin=parseFloat(amin);

 amax=readAttribute(thisElement,'max maximum amax ','');
 if (amax!=='') {
        if (isNaN(amax)) return {message:'Error: non-numeric value for amax attribute in '+thisElement+': '+amax,status:"error"};
 }
 amax=parseFloat(amax);


 aprec=readAttribute(thisElement,'prec round precision ','')
 if (aprec!=='') {
        if (isNaN(aprec)) return {message:'Error: non-numeric value for aprec attribute in '+thisElement+': '+aprec,status:"error"};
 }

 bok=readAttribute(thisElement,'blankok bok blank','0');

// build possible error message

  myvalue=get_elemValue(thisElement,adef); myvalue0=myvalue;


 if (trim(myvalue)==='' && bok=='1') return myvalue ;   // blank, and blank okay
 if (trim(myvalue)==='' && adef!==''){
    return adef ; // a default exists?
 }

    amess="Sorry, for "+thisElement+" you must enter a number (instead of ... "+myvalue0+") " ;
    if (isNaN(myvalue)) {             // remove commas?
      tt=''+myvalue ;
      tat1=tt;
      if (tt.indexOf(',')>-1)  tt=tt.replace(/\,/g,'');
       tt=trim(tt);
       if (isNaN(tt)) {
           a1=tt.substr(tt.length-1,1).toUpperCase();
           if (a1=='K') {
                tt=tt.substr(0,tt.length-1); if(!isNaN(tt)) tt=tt*1000;
           }
           if (a1=='M') {
                tt=tt.substr(0,tt.length-1); if(!isNaN(tt)) tt=tt*1000000;
           }
           myvalue=tt;
       } else {
           myvalue=tt ;
       }

 }


    if (isNaN(myvalue) || trim(myvalue)=='')  return {message:'  '+amess,status:"warning reset  quit"};
    myvalue=parseFloat(myvalue);

    if (amin!=='' && amax!=='')   amess=amess+' between '+amin+' and '+amax;
    if (amin=='' && amax!=='')   amess=amess+'  less than (or equal to) '+amax;
    if (amin!=='' && amax=='')   amess=amess+' more than (or equal to) '+amin ;

    if ( (amin!=='' && myvalue<amin) || (amax!=='' && myvalue>amax) )  return {message:' '+amess,status:"reset  quit warning"};


// if here, legit number (ignore amess)
    if (aprec!=='') {
        aprec=parseFloat(aprec);
        tt=myvalue/aprec;
        tt=tt.toFixed(0);
        myvalue=tt*aprec;
    }

    if (andec!=='') {                  // reformat?
        myvalue=myvalue.toFixed(andec);
    }

    return myvalue ;


 }                    // numeric check

//==============
// add commas to value (if numeric), add K and M, etc

function formatNumber(thisElement,changedElement,icall,opts0) {

   var myvalue,opts;
   var acomma=0, asK=0, asM=0, asCash=0,asDec=0,asPct=0,ndec=0,foo,foo2,itmp,ndec2,asInt=0,dotrim=0,dodebug=0,toUpper=0,toLower=0,noSpaces=0;


   if (opts0.length==0) opts0=['comma'];

   opts=opts0.join(' ');
   if (trim(opts)=='')       return {status:"nosave"} ;

   opts='xx '+opts.toUpperCase()+'  ';

       if (opts.indexOf(' DEBUG ')>0   ) dodebug=1;

    if (opts.indexOf(' TRIM ')>0   ) dotrim=1;
    if (opts.indexOf(' UPPER  ')>0   ) toUpper=1;
    if (opts.indexOf(' LOWER  ')>0   ) toLower=1;
    if (opts.indexOf(' NOSPACES  ')>0   ) noSpaces=1;

   if (opts.indexOf(' COMMA ')>0  || opts.indexOf('ADDCOMMA')>0) acomma=1;
   if (opts.indexOf(' M ')>0) asM=1;
   if (opts.indexOf(' INT ')>0  ||  opts.indexOf(' I ')>0 || opts.indexOf(' INTEGER ')>0 )  asInt=1;
   if (opts.indexOf(' K ')>0) asK=1;
   if (opts.indexOf(' PCTH ')>0 || opts.indexOf(' % ')>0 || opts.indexOf(' PERCENT ')>0 ) asPct=1;

   if (opts.indexOf(' CASH ')>0 || opts.indexOf(' $ ')>0) asCash=1;

   if (opts.indexOf(' DEC')>0) {
      itmp=opts.indexOf(' DEC');
      foo=opts.substr(itmp+4);
      foo2=getWord(foo,1);
      if (isNaN(foo2)) {
          asDec=0;
      } else {
         ndec=parseInt(foo2);
         asDec=1;
      }
   }
   myvalue=get_elemValue(thisElement,'');

   if (isNaN(myvalue) ) {                          // nan...  so do trim, lower, upper, etcmaybe trim
       if (dotrim==1)  myvalue=trim(myvalue);
       if (toUpper==1) myvalue=myvalue.toUpperCase();
       if (toLower==1) myvalue=myvalue.toLowerCase();
       if (noSpaces==1) myvalue=myvalue.replace(/\s/g,'');
       if (( dotrim+toUpper+toLower+noSpaces)==0)       return {status:"nosave"} ;
       return myvalue ;
    }

   myvalue=parseFloat(myvalue);

   if (asInt==1) {            // INT trumps DECn
      asDec=1 ; ndec=0 ;
   }

   if (asCash==1) {

     if (isNaN(myvalue)) return '...'       ;            // give up if nonnumeric
      var isneg=0;
      if (myvalue<0) isneg=1;
      myvalue=Math.abs(myvalue);
      ndec2=0;
      if (asDec==1) ndec2=Math.min(2,ndec);
      myvalue=parseFloat(myvalue.toFixed(ndec2));   // trim it

      if (acomma==1) myvalue=addComma(myvalue);

      if (isneg==0) {
            myvalue='$'+myvalue;
      } else {
            myvalue='-$'+myvalue ;
      }
      return myvalue ;
   }

    if (asPct==1) {
      myvalue=myvalue*100;
      myvalue=myvalue.toFixed(ndec2);
      if (acomma==1) myvalue=addComma(myvalue);
      myvalue=myvalue+'%';
      return myvalue ;
   }


   if (asDec==1) {
      myvalue=myvalue.toFixed(ndec);
      if (acomma==1) myvalue=addComma(myvalue);
      return myvalue ;
   }

   if (asM==1  && myvalue>1000000) {
       myvalue=myvalue.toFixed(0);
       myvalue=myvalue/1000000;
       myvalue=myvalue.toFixed(2);
       myvalue=addComma(myvalue);
       myvalue=myvalue+'m';
       return myvalue ;
   }

   if (asK==1 || asM==1) {              // if asM=1 and here, less than 1,000,000; so use x.xxk format
       myvalue=myvalue.toFixed(0);
       if (myvalue<1000) return myvalue;
       myvalue=myvalue/1000;
       myvalue=myvalue.toFixed(2);
       myvalue=addComma(myvalue);
       myvalue=myvalue+'k';
       return myvalue ;
   }

   if (acomma) {  // just acomma!
      myvalue=addComma(myvalue);
      return myvalue;
   }
   return myvalue ;
 }

//==============                               addcomma
// thisElement value = changedElement value.
// if changedElement is a number (i.e.; a 0 that would be used on initialization), no changes

function  useValue(thisElement,changedElement) {
  if (!isNaN(changedElement))   return {status:"skip"} ;
  var aval=get_elemValue(changedElement,'');
  return {status:"save",value:aval};
}

//===================
// show a message in an alert box -- used by alert as attribute . Note: read message from alert attribute (alist argument not used)
function showAnAlert1(thisElement,changedElement,icall,alist,origId) {
   if (origId==0) return {status:"nosave"};
   if (thisElement!==changedElement) return {status:"ok"} ;   // do NOT do an alert if changed by other element
   var goo,amess,alist ;
   amess=readAttribute(thisElement,'alert','');
   alist=[amess];
   goo=showAnAlert(thisElement,changedElement,icall,alist,origId) ;
   return goo
}

//===================
// show a message in an alert box -- used by alert as function (and called by alert as attribute

function showAnAlert(thisElement,changedElement,icall,alist,origId) {
   var reps=[],amess;

   amess=alist.join(',');
   amess='  '+trim(amess)+'  ';

   reps[0]={toFind:' this.id ',replaceWith:thisElement};
   reps[1]={toFind:' changed.id ',replaceWith:changedElement};

     aval=get_elemValue(thisElement,'');
   reps[2]={toFind:' this.value ',replaceWith:aval};

   if (typeof(watchData['flags'][changedElement])!=='undefined') {
      aval=watchData['flags'][changedElement] ;
   } else {
     aval=get_elemValue(changedElement,'');
   }
   reps[3]={toFind:' changed.value',replaceWith:aval};

   aval=readAttribute(thisElement,'default def ','');
   reps[4]={toFind:' this.default ',replaceWith:aval} ;

   amess=someHTMLDecode(amess,reps);
   alert(amess);
   return {status:"nosave"};
}

//======================
// confirm as attribute
//===================
// show a message in an confirm box -- used by alert as attribute . Note: read message from  (alist argument not used)
function showAnConfirm1(thisElement,changedElement,icall,alist,origId) {
   if (origId==0) return {status:"nosave"};
   if (thisElement!==changedElement) return {status:"ok"} ;   // do NOT do a confirm if changed by other element
   var goo,amess,alist,nowval,adef,aval ;
   amess=readAttribute(thisElement,'confirm','');
   alist=[amess];

   aval=get_elemValue(thisElement,'');
   adef=readAttribute(thisElement,'default def ',null);
   if (adef==aval)  return {status:"ok"};     //    if entered value is default, just accept it (no confirm done)

   goo=showAnConfirm(thisElement,changedElement,icall,alist,origId) ;
   return goo        ;
}

//===================
// show a message in an confirm box -- used by confirm as function (and called by confirm as attribute
// if "cancel", return status:'skip', else return status:'nosave'

function showAnConfirm(thisElement,changedElement,icall,alist,origId) {
   var reps=[],amess,adef,dodef,nowval;

   amess=alist.join(',');
   amess='  '+trim(amess)+'  ';

   reps[0]={toFind:' this.id ',replaceWith:thisElement};
   reps[1]={toFind:' changed.id ',replaceWith:changedElement};

     nowval=get_elemValue(thisElement,'');
   reps[2]={toFind:' this.value ',replaceWith:nowval};
     aval=get_elemValue(changedElement,'');
   reps[3]={toFind:' changed.value',replaceWith:aval};

   adef=readAttribute(thisElement,'default def ',null);
   dodef=0;
   if (adef==null) {
       aval='';
   } else {
       dodef=1;
       aval=adef;
   }
   reps[4]={toFind:' this.default ',replaceWith:aval};


   amess=someHTMLDecode(amess,reps);

   v1=confirm(amess);

   if (v1) return {status:"ok"};
   if (dodef==0) {
      return {status:"skip reset"} ;   // no default, use value before element was changed
   } else {                        // default exists, use it
      return adef ;      // replace with default value
   }

}



//===================
//  evaluate an expression and return what it returns

function doWatchEval(thisElement,changedElement,icall,alist,origId) {
   var reps=[],amess,adef,dodef,nowval,tt;
   if (origId==0) return {status:'nosave'} ;

   amess0=readAttribute(thisElement,'EVAL EVALUATE','');
   if (trim(amess0)=='')  abortJavaScript(' Error: '+thisElement+' contains an EVAL call, but no EVAL attribute ');

   amess='  '+trim(amess0)+'  ';

   reps[0]={toFind:' this.id ',replaceWith:thisElement};
   reps[1]={toFind:' changed.id ',replaceWith:changedElement};

     nowval=get_elemValue(thisElement,'');
   reps[2]={toFind:' this.value ',replaceWith:nowval};
     aval=get_elemValue(changedElement,'');
   reps[3]={toFind:' changed.value',replaceWith:aval};

   adef=readAttribute(thisElement,'default def ',null);
   dodef=0;
   if (adef==null) {
       aval='';
   } else {
       dodef=1;
       aval=adef;
   }
   reps[4]={toFind:' this.default ',replaceWith:aval};
   amess=someHTMLDecode(amess,reps,1);                 // don't do the &entities; decoding

  try {
    tt=eval(amess);
  }
  catch(err) {                          // possibly a comment  (anode.attributes.length doesnt exist)
    abortJavaScript('Error: in element '+thisElement+' an EVAL caused an error ('+err+'): '+amess0);
  }

   return tt ;


 }





//======================
// sum up all elements in an elements "watch" list.
// if any of the values (of these watched elements) is non-numeric, issue a warning and do NOT change
function sumWatch(thisElement,changedElement,icall) {
 var alist,asum,a1,v1,ndo,asum,aval;
  if (typeof(watchData['watch'][thisElement])=='undefined')   return {status:'skip'};

  alist=watchData['watch'][thisElement];
  asum=0;
  for (im=0;im<alist.length;im++) {
     a1=alist[im].ID;
     if (a1==thisElement) continue ;
     v1=get_elemValue(a1,'');
     if (trim(v1)=='' || isNaN(trim(v1)) ) {
       return {status:"warning skip", message:"Error in sumWatch in "+thisElement+": non-numeric value in element "+a1};
     }
     asum=asum+parseFloat(v1);
  }
  return asum;
}



//======================
// sum up all elements in an elements "watch" list.
// if any of the values (of these watched elements) is non-numeric, skip
function sumWatch0(thisElement,changedElement,icall,args,origId) {
 var alist,asum,a1,v1,ndo,asum,aval;
  if (typeof(watchData['watch'][thisElement])=='undefined')   return {status:'skip'} ;
  
  alist=watchData['watch'][thisElement];
  asum=0;
  for (im=0;im<alist.length;im++) {
     a1=alist[im].ID;
     if (a1==thisElement) continue ;
     v1=get_elemValue(a1,'');
     if (trim(v1)=='' || isNaN(v1) ) continue ;
     asum=asum+parseFloat(v1);
  }

  return asum;
}


//======================
// product of of all elements in an elements "watch" list.
// if any of the values (of these watched elements) is non-numeric, issue a warning and do NOT change
function prodWatch(thisElement,changedElement,icall) {
 var alist,asum,a1,v1,ndo,asum,aval;
  if (typeof(watchData['watch'][thisElement])=='undefined')   return {status:'skip'};

  alist=watchData['watch'][thisElement];
  asum=1;
  for (im=0;im<alist.length;im++) {
     a1=alist[im].ID;
     if (a1==thisElement) continue ;
     v1=get_elemValue(a1,'');
     if (trim(v1)=='' || isNaN(trim(v1)) ) {
       return {status:"warning skip", message:"Error in sumWatch in "+thisElement+": non-numeric value in element "+a1};
     }
     asum=asum*parseFloat(v1);
  }
  return asum;
}



//======================
// product of all elements in an elements "watch" list.
// if any of the values (of these watched elements) is non-numeric, skip
function prodWatch0(thisElement,changedElement,icall) {
 var alist,asum,a1,v1,ndo,asum,aval;
  if (typeof(watchData['watch'][thisElement])=='undefined')   return {status:'skip'} ;

  alist=watchData['watch'][thisElement];
  asum=1;
  for (im=0;im<alist.length;im++) {
     a1=alist[im].ID;
     if (a1==thisElement) continue ;
     v1=get_elemValue(a1,'');
     if (trim(v1)=='' || isNaN(trim(v1)) ) continue ;
     asum=asum*parseFloat(v1);
  }
  return asum;
}




//======================
// average of of all elements in an elements "watch" list.
// if any of the values (of these watched elements) is non-numeric, issue a warning and do NOT change
function avgWatch(thisElement,changedElement,icall) {
 var alist,asum,a1,v1,ndo,asum,aval;
  if (typeof(watchData['watch'][thisElement])=='undefined')   return {status:'skip'};

  alist=watchData['watch'][thisElement];
  asum=0;
  ndo=0;
  for (im=0;im<alist.length;im++) {
     a1=alist[im].ID;
     if (a1==thisElement) continue ;
     v1=get_elemValue(a1,'');
     if (trim(v1)=='' || isNaN(trim(v1)) ) {
       return {status:"warning skip", message:"Error in sumWatch in "+thisElement+": non-numeric value in element "+a1};
     }
     ndo++;
     asum=asum+parseFloat(v1);
  }
  aval=0;
  if (ndo>0)  aval=asum/ndo ;
  return aval;
}



//======================
// average of all elements in an elements "watch" list.
// if any of the values (of these watched elements) is non-numeric, skip
function avgWatch0(thisElement,changedElement,icall) {
 var alist,asum,a1,v1,ndo,asum,aval;
  if (typeof(watchData['watch'][thisElement])=='undefined')   return {status:'skip'} ;

  alist=watchData['watch'][thisElement];
  asum=1;
  for (im=0;im<alist.length;im++) {
     a1=alist[im].ID;
     if (a1==thisElement) continue ;
     v1=get_elemValue(a1,'');
     if (trim(v1)=='' || isNaN(trim(v1)) ) continue ;
     ndo++;
     asum=asum+parseFloat(v1);
  }
  aval=0;
  if (ndo>0)  aval=asum/ndo ;
  return aval;
}


//===========
// fade an element out of existence. This means showing it, and the slowly making it transparent.. and then hiding it
// if nsecs<0 ... fade IN (start with it off)
// if adiv is already being faded, stop fade and restart (maybe hide or display or turn on)

function fadeAnElement(thisElement,changedElement,icall,dlist,origId,cElem) {
  var ifoo,adiv,pstep,iat;
  var nsecs=5,totalTime=30,adiv='.',fadein=1;

  if (origId==0)  return {status:'nosave'};

 if (dlist.length>2)   nsecs=dlist[2] ;
 if (dlist.length>1)   totalTime=dlist[1] ;
 if (dlist.length>0) adiv=dlist[0];

 if (adiv=='.') adiv=thisElement ;
 adiv=trim(adiv);
 if (adiv=='..') {
     ifoo=document.getElementById(thisElement);
     adiv=ifoo.parentNode.id;
 }

if (typeof(watchData['timeouts'][adiv])!=='undefined' ) {
   clearTimeout(watchData['timeouts'][adiv]);     // an active timeout -- clear it
   watchData['timeouts'][adiv]=undefined;
}
ifoo=document.getElementById(adiv);
if (isNaN(totalTime) || totalTime<=0)   return {status:'nosave'};                           // totaltime=0 means "do nothing" (useful to clear a timeout)

if (isNaN(nsecs)) nsecs=totalTime ; // a sudden fadeout

nsecs=parseFloat(nsecs); totalTime=parseFloat(totalTime);
if (nsecs<0) {           // <0 means fadein
   fadein=-1;
   nsecs=-nsecs ;
}
if (nsecs>totalTime) nsecs=totalTime ; // a sudden fadeout / fadein

pstep=fadein*nsecs/totalTime ;

turnDisplayOn(adiv);                                 // turn it on
if (fadein>0) {
   iat=1.0;
   ifoo.style.opacity=1.0 ;
} else {
   iat=0.0;
   ifoo.style.opacity=0.0 ;
}

watchData['timeouts'][adiv]=setTimeout('fadeAnElement2("'+adiv+'",'+pstep+','+iat+','+nsecs+')',nsecs*1000) // div to change, change in opacity, current opacity, seconds to wait

return {status:'nosave'};
}

// helper for fadeAnElement
function fadeAnElement2(adiv,pstep,nowOpacity,nsecs) {
var ifoo,newOpacity;
ifoo=document.getElementById(adiv);
newOpacity=nowOpacity-pstep ;
if (newOpacity<0.00001) {                              // all gone. Reset opacity, and set to display:none
  ifoo.style.opacity=1.0;
  ifoo.style.display='none';
  watchData['timeouts'][adiv]=undefined ;             // timeout over
  return 1;                                         // all done
}
if (newOpacity>=1.0) {                              // all in.  set opacity to 1.0
  ifoo.style.opacity=1.0;
  watchData['timeouts'][adiv]=undefined ;             // timeout over
  return 1;                                         // all done
}
newOpacity=newOpacity.toFixed(2);
ifoo.style.opacity=newOpacity ;
watchData['timeouts'][adiv]=setTimeout('fadeAnElement2("'+adiv+'",'+pstep+','+newOpacity+','+nsecs+')',nsecs*1000) // div to change, change in opacity, current opacity, seconds to wait
return 1;
}


//========================
// turn on/off display of elements
// first search for matching id (with possible replaceiment of * with group attribute)
// if no such id, search for name (with possible * replacement) -- and apply changes to all that have this name
function doDisplayElement(thisElement,changedElement,icall,dlist,origId,cElem) {

 var disps,dlist,agroup,noDisp,doDisp,m,aa,aa1,bb,astyle,atag,a1,toggleDisp,nlist,inn;

 if (origId==0) return {status:"nosave"} ;               // do NOT do display elements on initialization
 if (dlist.length==0) return {status:"nosave"} ;        // no display attribute, so do nothing

 agroup=readAttribute(thisElement,'group','');
 noDisp=[];
 doDisp=[];
 toggleDisp=[];

 for (m=0;m<dlist.length;m++) {
     aa=dlist[m];

 // "group" idstring replacement?
     if (agroup!=='') aa=aa.replace('*',agroup);
     aa1=aa.substr(0,1);
     if (aa1=='-') {
        bb=aa.substr(1);
        if (bb=='1'){
           bb=thisElement ;
        }
        ibb=document.getElementById(bb);
        if (!ibb)  {
            nlist=document.getElementsByName(bb);
            if (nlist.length==0) return {status:"error",message:"Error:  element in (no) display list of "+thisElement+" does not exist: "+aa};
            for (inn=0;inn<nlist.length;inn++) noDisp.push(nlist[inn]);
        } else {
            noDisp.push(ibb);
        }
     } else if (aa1=='@') {
        bb=aa.substr(1);
        if (bb=='1') bb=thisElement ;
        ibb=document.getElementById(bb);
        if (!ibb)  {
            nlist=document.getElementsByName(bb);
            if (nlist.length==0) return {status:"error",message:"Error:  element in (toggle) display list of "+thisElement+" does not exist: "+aa};
            for (inn=0;inn<nlist.length;inn++) toggleDisp.push(nlist[inn]);
        } else {
            toggleDisp.push(ibb);
        }
     } else {
        if (aa=='1') aa=thisElement ;
        ibb=document.getElementById(aa);
        if (!ibb)  {
            nlist=document.getElementsByName(aa);
            if (nlist.length==0) {
               return {status:"error",message:"Error:  element in display list of "+thisElement+" does not exist: "+aa};
            }
            for (inn=0;inn<nlist.length;inn++) doDisp.push(nlist[inn]);
        } else {
            doDisp.push(ibb);
        }
     }
  }
 // first, no display everything in noDisp
 for (m=0;m<noDisp.length;m++) {
    a1=noDisp[m];
    turnDisplayOn(a1,'off',thisElement);
 }


// now, display everything in doDisp
 for (m=0;m<doDisp.length;m++) {
    a1=doDisp[m];
    turnDisplayOn(a1,'',thisElement);          // display element, based on type of element
 }

 // now, toggle display of everything in toggleDisp
 for (m=0;m<toggleDisp.length;m++) {
    a1=toggleDisp[m];
    turnDisplayOn(a1,'toggle',thisElement);          // display element, based on type of element
 }


 return {status:"nosave"}  ;   // no changes, do next function


}



//========================
// turn on/off display of elements conditional on refValue, etc
function noDisplayIf(thisElement,changedElement,icall,condlist) {
 var agroup,aref,acond,myvalue,aa,m,arefVal,nlist,aelem,donames,thisRefVal,ado ;
 if (changedElement==0) return {status:"nosave"} ;               // do NOT do display elements on initialization
 if (condlist.length==0) return {status:"nosave"} ;        // no display attribute, so do nothing
 agroup=readAttribute(thisElement,'group','');
 donames=trim(condlist[0]) ;
 if (agroup!=='') donames=donames.replace('*',agroup);
 aref='refValue';
 if (condlist.length>1 && trim(condlist[1])=='') aref=trim(condlist[1]);

 acond='LT';
 if (condlist.length>2 ) acond=trim(condlist[2].toUpperCase()) ;
 myvalue=get_elemNumber(thisElement,0);

 nlist=document.getElementsByName(donames);

 if (nlist.length==0) return {status:"nosave"} ;   // nothing to change, so done
 for (m=0;m<nlist.length;m++) {
     aelem=nlist[m];
     thisRefVal=parseFloat(readAttributeNode(aelem,aref,''));
     if (trim(thisRefVal)==='') continue ;

     ado='on';

     if (acond=='LT') {         // less than
        if (thisRefVal<myvalue) ado='off';
    }
     if (acond=='LE') {         // less than
        if (thisRefVal<=myvalue) ado='off';
    }
     if (acond=='EQ') {         // less than
        if (thisRefVal==myvalue) ado='off';
    }
     if (acond=='NE') {         // less than
        if (thisRefVal!==myvalue) ado='off';
    }
     if (acond=='GT') {         // less than
        if (thisRefVal>myvalue) ado='off';
    }
     if (acond=='GE') {         // less than
        if (thisRefVal>=myvalue) ado='off';
    }

    turnDisplayOn(aelem,ado,thisElement);
 }

 return {status:"nosave"}  ;   // no changes, do next function


}

//=================
// show if changed element in condlist
// origid=2 occurs only if save="2" and a load. Special case!
function showIf(thisElement,changedElement,icall,condlist,origId) {

   var useval,m,a1;
   if (origId=='2') {                                               // special case
            turnDisplayOn(thisElement,'',thisElement);
   }

    if (thisElement==changedElement && origId!='2') return {status:"nosave"};
    if (changedElement==0) return {status:"nosave"};

    useval=trim(get_elemValue(changedElement,''))  ;
    for (m=0;m<condlist.length;m++) {
        a1=trim(condlist[m]);
        if (a1==useval) {
// if (document.getElementById('debugNow').checked) alert(a1+' showif turning ont '+thisElement);
           turnDisplayOn(thisElement,'',thisElement);
           return {status:"nosave"};
        }
    }
    return {status:"nosave"};
}

//=================
// show if changed element in condlist
function hideIf(thisElement,changedElement,icall,condlist,origId) {
    var useval,m,a1;
    if (thisElement==changedElement) return {status:"nosave"};
    if (changedElement==0) return {status:"nosave"};
    useval=trim(get_elemValue(changedElement,'')) ;

    for (m=0;m<condlist.length;m++) {
        a1=trim(condlist[m]);
        if (a1==useval) {
// if (document.getElementById('debugNow').checked) alert(a1+' hidief turning OFF '+thisElement);
           turnDisplayOn(thisElement,'OFF',thisElement);
           return {status:"nosave"};
        }
    }
    return {status:"nosave"};
}



//========================================
// update a flag, returns nothing (nosave) -- or a space delimite list of flags
// if only one argument, the flag is thisElement (i.e.; called by a custom javasript function
// otherwise, called from a watchCall list : flag is from the flag attribute of thisElement

function updateFlag(thisElement,changedElement,icall,alist) {
  var aflag,goo,tmp, bflag,asec,bflagval,iscustom=0;

  goo=new Date();
  asec=goo.getTime();

  if (arguments.length<4) {                                 // explicit function call
        bflag=trim(thisElement) ;
        if (arguments.length>1) asec=changedElement;
        watchData['flags'][bflag]=asec;          // might be an explicit value
        return {status:"nosave"};
  }

// else, watchcall call
  if (changedElement==0)    return {status:"nosave"};

  aflag=readAttribute(thisElement,'flag','GENERIC_FLAG');
  aflag=trim(aflag);
   flist=getWord(aflag);
   for (im=0;im<flist.length;im++) {
      bflag=flist[im];
      if (bflag.indexOf('=')>-1) {
           tmp=bflag.split('=');     // for now, ignore any =xx stuff
           bflag=tmp[0];
           bflagval=tmp[1] ;
           if (watchData['flags'][bflag]==bflagval)  continue ;   // "turn on once" has already been turned on, so skip (note: could be seen as "not turned on yet" if set to a different value)
           asec=bflagval ;    // not set, set to explicit value
      }
      watchData['flags'][bflag]=asec;          // might be an explicit value
      watchData['flagsChanged'][bflag]=1;                                        // note on the list of flags changed
   }

  return {status:"nosave"};
}


//==========================
// set background color
// 3 args in color list
//  0: color to use if "changedElement" is neither thisElement nor altId
//  1: color to use if changedELement IS thisElement or altId
//  2: altId
function setBackgroundColor(thisElement,changedElement,icall,colorlist) {

  if (changedElement==0) return {status:"nosave"} ;               // do NOT do display elements on initialization
  var altId='';
  if (colorlist.length>2) altId=trim(colorlist[2]) ;
  var agroup=readAttribute(thisElement,'group','');
  if (agroup!=='') altId=altId.replace('*',agroup);

  if (typeof(changedElement)=='object') changedElement=changedElement.flagName;

  if (colorlist.length<2) return {status:"nosave"} ;        // no display attribute, so do nothing


  var iid=document.getElementById(thisElement);
  if (changedElement==thisElement || altId==changedElement) {                        // self (or alternate) change -- use colorlist[1]
    iid.style.backgroundColor=colorlist[1];
  } else {
   iid.style.backgroundColor=colorlist[0];
  }
  return {status:"nosave"} ;
}


// =================
// Basic call: if thisElement==changedElement, skip the next function.
// argumets
// First argument:
//    1 (the  default) : if thisElement==changedElement, skip the next functin (in the watchCall list)
//    0 : if thisElement!==changedElement, skip the next functin (in the watchCall list)
// Second argument:
//   Number of functions to skip (default is 1). If more than number remamining, same as a skip
//   Negative numbers and 0 cause an error
function jumpIfSelf(thisElement,changedElement,icall,alist,origId) {

   var skipSame,nskipSame,doskip;
  if (alist.length<1) {
     skipSame='1';
  } else {
     skipSame=trim(alist[0]);
  }
  if (alist.length<2) {
     nskipSame=1;
  } else {
     nskipSame=alist[1];
     if (isNaN(trim(nskipSame)) || trim(nskipSame)=='' || parseInt(nskipSame)<1) {
         abortJavaScript('Error in element '+thisElement+': bad jumpIfSelf argument: '+alist.join(' , ')) ;
      }
  }
  doskip=0;
  if (skipSame=='1' && thisElement==changedElement) doskip=1;
  if (skipSame!=='1' && thisElement!==changedElement) doskip=1;
  if (doskip==0)   return {status:"nosave" };
  nskipSame=parseInt(nskipSame);
  return {status:"jump",value:nskipSame}

}


//=============
// dummy watchCall -- does nothing. Sometimes this is useful, as it simulates a "no change" change (to catch attention of watchedBy elements
function doNothing(a) {
  return {status:'nosave'}
}

//=================
// do this watchCall list on initialization.
// called as: init(arg) (in watchCall list)
// If this is called during initialization, the value of arg will cause:
//    0 or NO (case insensitive) : SKIP  (i.e.; watchCall functions should NOT  be processed during initialization)
//    otherwise: return with nosave (so do next function)
// If this is called NOT during initialization, the value of arg will cause:
//    0 or NO (case insensitive) : do NOT skip
//    otherwise: skip (i.e.; watchCall functions should ONLY be processed during initialization)
//         otherwise, continue (nosave)
// The default value is "no"
function skipIfInit(thisElement,changedElement,icall,alist,origId) {
  var ainit,doIfInit=1,doit;

  if (alist.length==0) {
     doit='INIT';
  } else {
      doit=trim(alist[0]).toUpperCase() ;
  }
  if (doit!=='NOTINIT') doit='INIT';

  if (doit=='INIT') {
      if (origId=='0')  return {status:"nosave skip"} ;
  }
  if (doit=='NOTINIT') {
      if (origId!=='0')  return {status:"nosave skip"} ;
  }
  return {status:"nosave"};
}

//==================
// <alert> yileds a watchCall here
function doAnAlert(thisElement,changedElement,icall,alist,origId) {

  var onlist,offlist,lookfor,old1,itag,atag,ustyle;

  onlist=' '+readAttribute(thisElement,'onlist','')+' ';
  offlist=' '+readAttribute(thisElement,'offlist','')+' ';

  if (typeof(changedElement)=='object') {
     lookfor=' '+trim(changedElement.flagName)+' ';
  } else {
     lookfor=' '+changedElement+' ';
  }

  if (offlist.indexOf(lookfor)>-1) {
     turnDisplayOn(thisElement,'none',thisElement) ;
  }
  if (onlist.indexOf(lookfor)>-1) {
     turnDisplayOn(thisElement,'',thisElement) ;
     atag=readAttribute(thisElement,'alertButton','');
     if (atag!=='') {
         itag=document.getElementById(atag);
         if (itag){
            ustyle=readAttributeNode(itag,'onStyle','');
            if (ustyle!=='') {
               itag.setAttribute('style',ustyle);
               itag.style.cssStyle=ustyle;
            }
         }  // itag
     }   // atage
  }      // onlist

  return {status:"nosave"};
}    //alist


//====================================
// watchCall function: get update value and return it
function checkUpdates(thisElement,changedElement,icall,alist,origId) {
  var defval='';
  if (alist.length>0) defval=alist[0];
//  if (thisElement=='CyrMiles1_3') alert('CyrMiles1_3 getting update '+watchData['updates'][thisElement]);
  if (typeof(watchData['updates'][thisElement])=='undefined')  return defval;
  return watchData['updates'][thisElement] ;
}





//===========================
// some basic functions (augments those in utils.js)
//====================


// set element value and 'update'    : aval : stored in element, aval2: stored in 'updates'
// if aval null, jus store aval2 in updates
function update_elemValue(aelem,aval,aval2) {
  if (arguments.length<3) aval2=aval;
  var aelemId=aelem ;
  if (typeof(aelem)!=='string') aelemId=aelem.id ;
  set_elemValue(aelem,aval);
  saveUpdateWatch(aelemId,aval2);
  return 1;
}

// ====== save to 'updates'
function saveUpdateWatch(ado,aval,igroup) {
  if (arguments.length<3) {
    watchData['updates'][ado]=aval;
  } else {
    watchData['updates'][ado+'_'+igroup]=aval;
  }
  return 1;
}

// ====== read from 'updates'
function readUpdateWatch(ado,adef,igroup) {
  if (arguments.length<2) adef='' ;
  var aval ;
  if (arguments.length<3) {
     if (typeof(watchData['updates'][ado])=='undefined') {
         aval=adef;
     } else {
        aval=watchData['updates'][ado] ;
     }

  } else {
     if (typeof(watchData['updates'][ado+'_'+igroup])=='undefined') {
         aval=adef;
     } else {
        aval=watchData['updates'][ado+'_'+igroup] ;
     }
  }
  return aval ;
}


//============
// save all the specs -- this puts stuff into a new window that one can then save
function doSaveSpecs(i) {
var goo,alist,atime,oof,im,aval,amess,delays,aline;
amess=prompt('Enter a short description of the current set of parameters ');

if (amess==null ) return 0;
atime=get_currentTime(1,1);
goo='// carcosts parameters saved (after escaping) at '+atime+'\n// '+amess+'\n';

alert('A new tab/window will be opened. You should save its contents to a file (which can later be used by the Load button)','');

alist=document.getElementsByTagName('*');
delays='';
for (im=0;im<alist.length;im++) {
   oof=trim(readAttributeNode(alist[im],'save',''));
   if (oof=='') continue ;
   oof=' '+trim(oof.toUpperCase())+' ';
   aval='';
   if (oof.indexOf(' V')>-1)  aval=escape(get_elemValue(alist[im],''));
   aline=oof+'='+alist[im].id+'='+aval;

  if (oof.indexOf(' D')>-1)    {
     delays=delays+'\n'+aline ;
   } else {
     goo=goo+'\n'+aline ;
   }
}
goo=goo+delays;
displayStringInWindow(goo,0,2) ;
return 1 ;
}

//---------
// resetore parameters  saved in a file saved using the save button
function doRestoreSpecs(i) {
  var v0,v1,v2,vt,vc,vd,vi,vic,gothtml5=0,vtt;

if (window.File && window.FileReader && window.FileList && window.Blob) gothtml5=1 // html5 ?

v2=document.createElement('div');
v2.id="doRestoreSpecs_fileRead";
v2.class='loadParamsBox';
v2.setAttribute('class','loadParamsBox');

vc=document.createElement('input');
vc.type='button';
vc.value='x';
vc.title='Cancel (do not load parameters)';

vc.onclick=doRestoreSpecs2a ;

//addEvent(vc, 'click', doRestoreSpecs2a);
v2.appendChild(vc);
vt=document.createElement('b');
vtt=document.createTextNode('Restore parameters (that were previously saved using the Save button');
vt.style.paddingLeft="20px";
vt.appendChild(vtt);
v2.appendChild(vt);

vd=document.createElement('div');
vd.innerHTML='Although you can directly enter a list of parameters (and; values) to restore in the  text box, we recommend:'+
'<ol type="a">'+
'<li>Opening a parameters files, that you saved with the Save button, in your favorite text editor'+
'<li>Copy and paste contents of this file to the text box'+
'</ol>';
if (gothtml5==1) {
  vd.innerHTML=vd.innerHTML+'<em>or</em> ...</p> <ol type="i"> <li>Use the Browse button to select a file -- its contents will be automatically copied to the text box </ol> ';
}
vd.innerHTML=vd.innerHTML+' Regardless of how you enter information into the text box, once you are ready you should click here ... ';
v2.appendChild(vd);

// if no html5 support, user has to cut and paste


v0=document.createElement('input');
v0.type='button';
v0.value='Click to restore the parameters ';
v0.title='After retrieving (or pasting) contents of a saved parameters file into the text box, click here to use the specified values ';
v0.onclick=doRestoreSpecs2 ;
vd.appendChild(v0);

v2.appendChild(vd);

vt=document.createElement('p');
v2.appendChild(vt);

// if  html5 support, allow for fileread
if (gothtml5==1) {
  vi=document.createElement('input');
  vi.type='file';
  vi.title="You can find a file and copy its contents directly into the box ...";
  vi.id='doRestoreSpecs_input';
  vi.name='files[]';
  vi.onchange=doRestoreSpecs_a1;        // this is called when a file is chosen
  vi2=document.createElement('div');
  vi2.id='doRestoreSpecs_showFiles';
  v2.appendChild(vi);
  v2.appendChild(vi2);
}


 // always allow for cut and paste
v1=document.createElement('textarea');
v1.id='_LoadParamsBox_';
v1.setAttribute('rows',22)  ;
v1.setAttribute('cols',122);

if (gothtml5==0) {
  v1.title='Paste the contents of a saved-parameters file into here, and then click the `Click to load the parameters button` ' ;
} else {
  v1.title='Use the browse button to copy a previously saved parameters to this box... or you can manually paste the contents of a saved-parameters file into here. In either case, after copying information to here, please  click the `Click to load the parameters button` ' ;
}

v2.appendChild(v1);

document.body.appendChild(v2);

return 1;
}

//--------------------------
// onchange handler for fileread in doRestoreSpecs
function doRestoreSpecs_a1(evt) {

 var targ,files,output,i,f,afile,contents;

 if (!evt) var evt = window.event; // from     http://www.quirksmode.org/js/events_properties.html
 if (evt.target) targ = evt.target;
 else if (evt.srcElement) targ = evt.srcElement;
 if (targ.nodeType == 3) targ = targ.parentNode; //safari bug

 files=targ.files;    // files is a FileList of File objects. List some properties.
 output = [];
 for (var i = 0, f; f = files[i]; i++) { afile=f;                             // file is "last file in list". If not multiple, the only file.
      output.push('<li><strong>', escape(f.name),'</strong> (', f.type || 'n/a', ') - ',
                  f.size, ' bytes, last modified: ',
                  f.lastModifiedDate ? f.lastModifiedDate.toLocaleDateString() : 'n/a',
                  '</li>');
 }
 document.getElementById('doRestoreSpecs_showFiles').innerHTML = '<ul>' + output.join('') + '</ul>';   // display info on chosen file

 var reader = new FileReader();

 reader.readAsText(afile);

 reader.onload = function(evt2) {                             //  after file is retrieved this is called
   if (!evt2) var evt2 = window.event;
   if (evt2.target) targ2 = evt2.target;
   else if (evt2.srcElement) targ = evt2.srcElement;
   if (targ2.nodeType == 3) targ2 = targ2.parentNode;
   var i3=document.getElementById('_LoadParamsBox_');
   i3.value= targ2.result;                                         // fill in textarea
   return 1;


};

reader.onerror = function(evt2) {                 // on error this is called
   if (!evt2) var evt2 = window.event;
   if (evt2.target) targ2 = evt2.target;
   else if (evt2.srcElement) targ = evt2.srcElement;
   if (targ2.nodeType == 3) targ2 = targ2.parentNode;
   var goo='' ;
   for (a in targ2.error) goo=goo+''+a+'='+targ2.error[a]+'\n';
    alert("File "+afile.name+" could not be read! Code " +goo) ; //targ2.error.code );
};

 return 1;

}


function init1(i){
 document.getElementById('listDirContents').onchange=doListDir ;
}




//==========
// event function for doRestoreSpecs
function doRestoreSpecs2(e) {
var ia,stuff,lines,stuff,ndo,nmiss,nok,goo,im,aline,as,aid,avalue,iid,stat,asave,vstat,vstat2,alist;
var saveValue=0,saveWatch=0,saveCall=0,saveWatchedBy=0;

ia=document.getElementById('_LoadParamsBox_');
ndo=0; nmiss=0; nok=0;    goo='',goo2='';
stuff=ia.value;

ia=document.getElementById('doRestoreSpecs_fileRead');
ia.parentNode.removeChild(ia);

lines=stuff.split('\n');
for (im=0;im<lines.length;im++) {
  saveValue=0;saveWatch=0;saveCall=0;saveWatchedBy=0;

  aline=trim(lines[im]);
  if (aline.substr(0,2)=='//' || aline=='') continue;
  ndo++;
  a2=aline.split('=');

  aid=a2[1];

  asave=trim(a2[0].toUpperCase());
  asave=' '+asave+' ';

  if (asave.indexOf(' V')>-1) saveValue=1;
  if (asave.indexOf(' W')>-1) saveWatch=1;
  if (asave.indexOf(' B')>-1) saveWatchedBy=1;
  if (asave.indexOf(' C')>-1) saveCall=1;
  if (asave.indexOf(' A')>-1) {saveCall=1;saveValue=1;saveWatchedBy=1 ; }


  avalue=unescape(a2[2]);
  iid=document.getElementById(aid);
  if (!iid) {
     goo=goo+'\n Line'+im+': No such element (with id= '+aid+')  :: '+aline;
     nmiss++;
     continue ;
  }
  if (saveValue==1) {            // 2 means "just doWatch"

    stat=update_elemValue(iid,avalue);
    if (stat==0) {
      goo=goo+'\n Line '+im+': Unable to save value of '+aid+'  :: '+aline;
      continue ;
    }
  }
  if (saveCall==1) {                    // call this elements functions (as specified in its watchCall list). thisElement and changedElement both = thisElement
      vstat=doWatch1(aid,aid,1)  ;
  }

  if (saveWatch==1) {               // has any of what this element watching changed? If so, process this element (using changed element as changedElement
     if (typeof(watchData['watch'][aid])!=='undefined')  {
       alist=watchData['watch'][aid] ;
       for (iv=0;iv<alist.length;iv++) {
            if (alist[iv]!==aid) {
                vstat2=doWatch1(aid,alist[iv].ID,1) ;  // simulate change to first non-self element that is watching
                break ;
           }
       }
     }
   }

  if (saveWatchedBy==1) {               // process elements that are "watching" this element    (changedElement is this element, for each of the "watching" elements)
     vstat2=doWatch(aid,2000,0,1);
  }

  goo2=goo2+'\n '+aid+'== '+vstat;


}
if (goo!=='') {
   alert('For '+ndo+' parameters, '+nmiss+' could not be reset: \n'+goo);
} else {
   alert(ndo+' parameters were reset '+goo2);
}




return 1;
}

//==========
// event function for doRestoreSpecs
function doRestoreSpecs2a(e) {
var ia;

ia=document.getElementById('doRestoreSpecs_fileRead');
ia.parentNode.removeChild(ia);
return 1;
}
