/* library of proceudre for creating "nonParametric" probablity distribution function (pdf)

Copyright 2013. Daniel Hellerstein (danielh@crosslink.net)

This file is part of carCosts2.

    carCosts2 is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    carCosts2 is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with carCosts2.  If not, see <http://www.gnu.org/licenses/>.


 An "emprical" pdf is non-parametric, and is explicitily defined over integers between a minimum and maximum value
 However, marginal probability and cumulative probability values for any real number in this range can be computed

 These functions create, and work with, what are called "empircal PDF" objects (pdfObjs). A pdfObj contains
 the following fields:
      pdf: an array containing "emprical pdf" probability values. It is defined between for every integer between min and max. It's elements will sum to 1.0
      min : minimum value (the first defined element of the pdf array) and
      max: maximum value (the last defiend element of the pdf array)
      mean : mean value from this pdf
      cdf: cumulative distribution function corresponding to the points in the pdf (
 Note that the the pdf array can contain 0s (for zero probablity) anywhere, epecially at the either tails.
 Also, cdf[min]=0 and cdf[max]=1.0

  Functions:
        pObj=makeEmpricalPdf(min,max,xRef,probRef)   -- create an emprical pdf object
        pObjA=changeMeanPdf(pObj,newMean)        -- change mean of an empirical pdf, retaining upper and lower bounds
        pval=figPdfValue(x,pdfObj)                   -- marginal probality of x, using pdf defined in pdfObj (x can be any real number )
        cumval=figCdfValue(x,pdfObj)    -- cumulative probability of x, using pdf defined in pdfObj (x can be any real number )
        histoPdf(anid,pobj,height)   -- create a histogram of the pdf in the pobj emprical pdf object

 pObj=makeEmpricalPdf(min,max,xRef,probRef)
     Create a empircial PDF and
         min: minimum value (bottom of range) -- will be converred to an integer
         max: maximum value (top of range)  -- will be converred to an integer
         xref: array of "reference" values used to create the empirical pdf. Values in xref MUST be increasing: xref[n] < xref[n+1]
         probRef: probability associated with each point in xref. Values in probRef must be non-negative

     The xref and probRef are used to create probability values between min and max. Conceptually, you can think of a line plot using
     xref,probRef points; and then stretching this to cover the range betwen min and max. And thenm, these values are normalizzed
     so that the sum (over all the integers inclusive of min and max) sums to 1.0

     Conceptually, this means that the pdf created will be piecewise linear.

     Thus, the magnitude of the xref values, and the probref values, don't really matter -- both will be "normallized".
     The simplest example is a 2 element xref with 2 elements in probRef -- this creates a linear pdf. The value of xref
     wouldn't matter, and all that matters with the probRef values is their relative size (the values of probRef do NOT have
     to sum to 1.0).

     By choosing the xref values you can specify regions of the pdf that are jagged, and regions that are smooth.


 newPdfObj=changeMeanPdf(pdfobj,newmean)

    Modify a pdf, as contained in the pdfObj, to have a new mean while retaining its min and max.

      pdfObj : a pdf object (as produced by makeEmpiricalPdf
      newmean : the mean of the new pdf

      newPdfObj : a new pdf object, with a pdf array that will have a mean close to newmean

   Creating a pdf with a new mean, but the same lower and upper bounds, is tricky. Unless the old pdf has sufficient 0s in the tails,
   some kind of distortion is necessary. shiftEmprical uses as "squeezing" that maintains the overall shape of the distribution, but shortens
   its range of support. This means that one of the tails will have a number (perhaps a large number) of elements with 0 probablity

   For example, if you define a triangular distribution that starts at 0 and ends at 100, and peaks at 50 -- hence has a mean of fifty.
   You use changeMeanPdf to transform this distriution so it has a mean of 60.
   Then, elements 0 to 20 will have values of 0, and a new (steeper) triangular disitribution will be created between 20 and 100.

   This squeezing will not occur if there are sufficient zeros in a "tail".

   For example;
         xref=[100,120,130,140,150,160,170,180,190,2000]
         probRefs=[0,    0, 10,  0, 10,  0, 10, 0,  10,  0 ]
         pdfObj=makeEmpricalPdf(100,200,xref,probRefs) ;  // in this case the bounds of xref happen to be min and max
         pdfObjA=changeMeanPdf(pdfObj,50)

     pdfObj.pdf would be defined between 100 and 200; with 0 probablities between 0 and 120, and a sawtooth distribution between 120 and 200 with a mean of 60

     pdfObjA.pdf would also be defined between 100 and 200; with 0 probablities between 100 and 110, the same sawtooth between 110 and  190, and zeros between
     190 and 200

 newPdfObj=changeRangePdf(pdfobj,newmin,newmax)

    Modify a pdf, as contained in the pdfObj, to have a new min and max. T

      pdfObj : a pdf object (as produced by makeEmpiricalPdf
      min, max : the min and max of the new pdf

      newPdfObj : a new pdf object, with a pdf array that will have a min and max chosen

   Note that the mean of this is indeterminante. You may wish to call changeMeanPdf after changeRangePdf!


 pval=figPdfValue(x,pdfObj)

   Return a probability, given a PDF and a "x" value
      x : the value to find the   probablity of. Can be any real numbef betwen min and max
     pdfObj: pdf object, as created by makeEmpiricalPdf

    Rules:
     if x<min : return 0
     if x >max : return 0
     if x is an integer, return apdf[x]
     otherwise, return an linearly interpolated value

 cumval=figCdfValue(x,pdfObj)

   Return a cumulative probability, given a CDF and a "x" value
     x : the value to find the cumulative probablity of
     pdfObj: pdf object, as created by makeEmpiricalPdf

    Rules:
       if x<amin : return 0
       if x >amax : return 1
       if x is an integer, return acdf[x]
       otherwise, return an linearly interpolated value

histoPdf(anid,pobj,opts,xlabels,ylabels)

   Makes a histogram of an empirical pdf

   This used a <table> and css style elements. It isn't real elegant, but should be robust.

     anid:: id of container to put the <table> containing the histogram
     pobj: empricialPDF object.
     opts: space delimited list of options
     xlabel: list of where to put labels on the x axis

    The bars represent the probablity value. The heights of the bars are normallized by the largest probablity value (so at least one bar will be height pixels high)

     An mouseover title will be on each bar, that shows its x and probablity value. Also, the bar at the mean value will be colored lime (all other bars are colored red)

     Tthe following options are supported. The syntax of opts is:  "optname=optvalue optname2=optvalue2 ..."
     Note there shouuld NOT be a space between the optname, =, and optvalue. Also, do NOT use " or ' around optvalue.
        height=nnn         default is height=200      Maximum height of histogram bar, in pixels (not %).
        width=nn          default is width=10        Width of histogram bar, in pixels (not %)
        color=acolor       default is color=red.        Color of histogram bar. Note that color can be a #rrggbb "hex" color, or a color name
        meanColor=acolor   default is color=lime        Color of histogram bar at the mean value
        medianColor=acolor default is color=red         Color of the median value
        hiliteColor=acolor  default is cyan             Color of highlighted bar
        hilite=m1;m2;            default is ''          Bars to highlight (using hiliteColor). m1... refers to x values. Seperate with semi-colons -- NO SPACES!
        padding=nn        default is padding=1         Gaps between histogram bars, in pixels. padding=0 means "no gap"
        lineOnly=0/1        default is lineOnly=0       If 0, only the line (a dotted line) is shown. If 1, line and bar is shown (if lineColor='', line is NOT shown)
        lineColor=acolor    default is ''                Color of line (on top of histogram). Default is no line
        clear=1             value doesn't matter         If clear=1 is specified, contents of anid are cleared before historgram is added.
        class=aclass       default is class=tdHistoUp   Note that width and the 3 color values will override whatever is specified in class
    If median and mean are the same, mean color is used.

    xlabel is optional. If specified, it should be an array. The array should contain indices where a tic on an x axis should be placed.
        If xlabel is specified, two rows will be added: one with tic marks are the x values in xlabel, and another row beneath it with the x value

        Special: instead of a value, you can put a nn% -- which will be converted to a value using the min and max of the pdfObject.

        Examples:  histoPdf('myshow',pdfobj,' width=12 ',[5,10,50,100])
               histoPdf('myshow',pdfobj,' width=12 ',['10%','33%','66%','90%')

        You can mix and match numbers and 'nn%', but be careful. The numbers (after % conversion) MUST be in order ( [5,10,8,20] will cause odd problems)

    ylabel is optional. If specified, it should be an array. The array should contain where tics/labels on the y axis should be plades.
       These MUST be expressed as percentages (of the maximum y value). You can use a 'nn%', or must nn
       The % is the pct of the maximum y value. The axis will display a 3 digit precision number corresponding to this.
       So, if  the maximum y value is 0.50; then ylabel=[20,50,80] would yield labels of "0.10", "0.250", and "0.40" (appropriately located on the y axis)


   newObj=weightedSumPdf(pdfobjs,weights,donorm) {

    Create an array of weighted sums of several pdfs: each elment being the weighted sum of corresponding elements in the pdfs

       pdfobjs: array of "empirical pdf objects"
       weights: array of scalar weights
       donorm: 1 to return as a pdf (normalizzed, etc). Otherwise, just return a weighted sum (in the pdf field)

    Returns a pdf object (if donorm!=1, cdf and mean elements are NOT in this pdf object)

    Notes:
      * If donorm=1  pdf (normalized weighted sum), cdf, mean, min and max.
              otherwise: pdf (the weighted sum), min and max
      * min and max max are the lowest min and hightest max of the pdfs
      * The ranges of the several pdf objects do not have to overlap
        This means vales in the "created" pdf ares composed only from input pdfs that cover this index value
      * A non-normalized "pdf object" can be used as input to histoPdf.

==================================== */
// create empirical pdf object

function makeEmpiricalPdf(minval,maxval,refs,probs,newmean){

var m,newPdfs,newRange,oldMin,oldRange,oldMax,i0,o1,n1,n2,nrange,op1,op2,opRange,opdelta,jj,jm,psum,j,amean,cumc,newCdfs ;

 // make sure refs is legit

 if (refs.length<2) {
    alert('Error in makeEmpricalPdf: refs must be at least 2 elements long ');
    return null;
 }
 if (refs.length!==probs.length) {
    alert('Error in makeEmpricalPdf: refs has '+refs.length+' elements, but probs has '+probs.length+' elements.');
    return null;
 }
 for (var m=0;m<refs.length-1;m++) {
  m1=m+1;
  if (refs[m]>=refs[m1]) {
    alert('Error in makeEmpricalPdf: not strictly increasing refs (index '+m+'= '+refs[m]+ ', index '+m1+' =  '+refs[m1]);
    return null ;
  }
  if (probs[m]<0) {
    alert('Error in makeEmpricalPdf: probs element '+m+' is < 0 ('+probs[m]+')');
    return null ;
  }
}

 maxval=parseInt(maxval); minval=parseInt(minval);
 if (minval>=maxval) {
    alert('Error in makeEmpricalPdf: minval ('+minval+') is not greater than maxval ('+maxval+')');
    return null ;
  }

if (arguments.length>4) {
 if (newmean <= minval || newmean>=maxval) {
    alert('Error in makeEmpricalPdf: newMean ('+newmean+')is either le than minval ('+minval+') or ge than maxval('+maxval+')');
    return null ;
  }
}


// arguments are legit, so proceed

 newPdfs=[];
 for (var m=minval;m<=maxval;m++) newPdfs[m]=0;

 newRange=maxval-minval;
 oldMin=refs[0]; oldMax=refs[refs.length-1] ;
 oldRange=oldMax-oldMin;

 for (i0=0;i0<refs.length-1;i0++) {
    o1=refs[i0]; o2=refs[i0+1]; oRange=o2-o1 ;
    n1=minval + (((o1-oldMin)/oldRange)*newRange);
     n1=parseInt(n1);
    n2=minval + (((o2-oldMin)/oldRange)*newRange) ;
    n2=parseInt(n2);
    nrange=n2-n1 ;
    op1=probs[i0] ; op2=probs[i0+1]; opRange=op2-op1 ;
    opdelta=opRange/nrange ;
//    alert(' i0 '+i0+' with delta '+opdelta+' in n1 to n2 '+n1+' - '+n2);
    for (var jj=0;jj<=nrange;jj++) {
        jm=n1+jj ;
        aprob=op1+ (jj* opdelta) ;
        newPdfs[jm]=aprob;
    }
  }
  psum=0;

  for (var j=minval;j<=maxval;j++) {
     psum=psum+newPdfs[j] ; // normalize (force sum of probs =1 )
  }
  for (var j=minval;j<=maxval;j++) newPdfs[j]=newPdfs[j]/psum ;

// compute mean and cdf
 amean=0; cumc=0; newCdfs=[];
  for (var j=minval;j<=maxval;j++) {
    amean=amean+(j*newPdfs[j]);
    cumc=cumc+newPdfs[j];
    newCdfs[j]=cumc;
  }


//alert(newPdfs.join(' , '));
  return {pdf:newPdfs,cdf:newCdfs,min:minval,max:maxval,mean:amean};
}

//==============================
// squoosh and shift an emprical pdf: so that mean is newMean, but the upper and lower range is the same (so the range of support does not change)

function changeMeanPdf(pdfobj,newmean) {

var apdf,acdf,maxval,minval,amean,mdiff,mdiffA,tmp,ctmp,tmp2,j,useMax,useMin,zleft,zright;
var amean,csum,j,useLen,iclip,clipPct,keepPct,useLen2,c1,tmp2,tmean,asum,goo,x,c2,tmean,tmean2,nowmove,j3;

var useMin=-1,useMax=-1;

apdf=pdfobj.pdf ;acdf=pdfobj.cdf ; maxval=pdfobj.max  ; minval=pdfobj.min  ; amean=pdfobj.mean ;

mdiff=(newmean-amean).toFixed(0); ;
newmean=parseInt(newmean); amean=parseInt(amean);       // differences btween current and desired mean

if (mdiff==0) return {pdf:apdf,min:minval,max:maxval,mean:newmean,cdf:acdf}; // no diference,return results!

mdiffA=Math.abs(mdiff);

tmp=[];  ctmp=[]; tmp2=[];
for (var j=minval;j<=maxval;j++) {   // find true range of support
   tmp[j]=0 ; ctmp[j]=0;tmp2[j]=0 ;                              // also initialize
   if (apdf[j]>0) {
     useMax=j;
     if (useMin<0) useMin=j;
   }
}
if (useMax<0) {
   alert('Error: no non-zero probabliites in changeMeanPdf ');
   return null;
}
zleft=useMin-minval ;                     // zeros on left
zright=maxval-useMax ;                    // zeros on right
//alert(useMin+' '+useMax+' so ' +zleft+','+zright);

if ( (mdiff>0 && zright>mdiffA) || (mdiff<0 && zleft>mdiffA) ) {  // no need to squoosh, just shift

    mdiff=parseInt(mdiff);
    for (var j=minval; j<=maxval;j++) {
       j2=j+mdiff ;
       if (j2<minval || j2>maxval) continue ;
       tmp[j2]=apdf[j];
    }
    csum=0; amean=0;
    for (var j=minval;j<=maxval;j++)  {
      csum=csum+tmp[j]; ctmp[j]=csum;
      amean=amean+ (tmp[j]*j);
   }
   return {pdf:tmp,min:minval,max:maxval,mean:amean,cdf:ctmp}; // no diference,return results!
}                                           // end of no squoosh needed
// have to squoosh it
useLen=useMax-useMin ;
//alert(useMin+' - '+useMax+' : useLen '+useLen+','+mdiff );
if (mdiff>0) {
   iclip= mdiffA-zright;
   clipPct=iclip/ (useMax-amean);
} else {
   iclip= mdiffA-zleft;
   clipPct=iclip/ (amean-useMin);
}

keepPct=1-clipPct ;
useLen2=parseInt(keepPct*useLen);
//alert('iclip '+iclip + ' : '+keepPct);
cdo=useLen/useLen2;

c1=0;
tmp2=[];
tmean=0;  asum=0;

goo='';
for (var j=0;j<=useLen2;j++) {
    x=useMin+(j*cdo) ;
    c2=figCdfValue(x,pdfobj);
    tmp2[j]=c2-c1;
    asum=asum+tmp2[j];
    c1=c2;
}

tmean=0;
for (var j=0;j<=useLen2;j++) tmp2[j]=tmp2[j]/asum;
for (var j=0 ; j<=useLen2;j++) tmean=tmean+(tmp2[j]*j);

tmean2=minval+tmean;

nowmove=parseInt(newmean- tmean) ;
if (nowmove<0) {
   alert('Programming error in changeMeanPdf: nowmove <0 ');
   return null;
}
//alert(' uselen2 nowmove '+useLen2+','+nowmove);
for (var j=0;j<useLen2;j++) {
    j3=j+nowmove;
    if (j3>maxval) {
        alert('Programming eror in changeMeanPdf: j3> maxval ');
        return null
    }
    tmp[j3]=tmp2[j];
}
csum=0; amean=0; ctmp=[];
for (var j=minval;j<=maxval;j++)  {
    csum=csum+tmp[j]; ctmp[j]=csum;
    amean=amean+ (tmp[j]*j);
}
return {pdf:tmp,min:minval,max:maxval,mean:amean,cdf:ctmp}; // no diference,return results!


}

//===============
// Return a cumulative probability, given a CDF and a "x" value


function figCdfValue(x,pdfObj) {
var ix,ix1,v1,v2,v3,amin,amax,acdfs;
amin=pdfObj.min; amax=pdfObj.max; acdfs=pdfObj.cdf ;
 if (x<amin) return 0.0 ;
 if (x>=amax) return 1.0;
 ix=parseInt(x);
 if (typeof(acdfs[ix])=='undefined') {
      alert('Error: figCdfValue. Undefined element in acdf array (at: '+x+')');
      return null;
 }
 if (ix==x) return acdfs[ix];

 ix1=ix+1;
 if (typeof(acdfs[ix1])=='undefined') {
      alert('Error: figCdfValue. Undefined next element in acdf array (at '+ix1+')');
      return null;
 }

 v1=acdfs[ix]; v2=acdfs[ix1];
 if (v1>v2 || v1<0 || v2<0 || v2.toFixed(4)>1.0 ) { // check for wierd errors
      alert('Error: figCdfValue. Improper cdf using '+ix+'('+v1+') and '+ix1+ ' ('+v2+') ');
      return null;
 }

 dx=(v2-v1)*(x-ix);
 v3=v1+dx;
 return v3;
}


//===============
// Return a probability, given a PDF and a "x" value

function figPdfValue(x,pdfObj) {
var ix,ix1,v1,v2,v3,amin,amax,apdfs;
 amin=pdfObj.min ; amax=pdfObj.max ; apdfs=pdfObj.pdf ;

 if (x<amin) return 0 ;
 if (x>amax) return 0;
 ix=parseInt(x);
 if (typeof(apdfs[ix])=='undefined') {
      alert('Error: figPdfValue. Undefined element in apdf array (at '+ix+')');
      return null;
 }
 if (ix==x) return apdfs[ix];

 ix1=ix+1;
 if (typeof(apdfs[ix1])=='undefined') {
      alert('Error: figPdfValue. Undefined element in apdf array (at '+ix1+')');
      return null;
 }

 v1=apdfs[ix]; v2=apdfs[ix1];
 if (v1<0 || v2<0 ) { // check for wierd errors
      alert('Error: figPdfValue. Improper pdf between '+ix+' ('+v1+') and '+ix1+ ' ('+v2+') ');
      return null;
 }

 dx=(v2-v1)*(x-ix);
 v3=v1+dx;
 return v3 ;

}

//=========================
// make a histogram. Uses a table
//  where: id of container to put histogram in
//  pobj: empricialPDF object, or similar. Must have pdf, min, and max elements, can have mean and cdfs (if mean and cdfs doesn't exist, less info in bar by bar mouseovers)
// height: height, in pixels
// space delimited list of options
//        height=nn         default is height=200      Maximum height of histogram bar, in pixels (not %).
//        width=nn          default is width=10        Width of histogram bar, in pixels (not %)
//        color=acolor       default is color=red.        Color of histogram bar. Note that color can be a #rrggbb "hex" color, or a color name
//        hiliteColor=acolor  default is cyan             Color of highlighted bar
//        hilite=m1;m2;            default is ''          Bars to highlight (using hiliteColor). m1... refers to x values. Seperate with semi-colons -- NO SPACES!
//        meanColor=acolor   default is color=lime        Color of histogram bar at the mean value
//        medianColor=acolor default is color=red         Color of the median value
//        padding=nn        default is padding=1         Gaps between histogram bars, in pixels. padding=0 means "no gap"
//        lineOnly=0/1        default is lineOnly=0       If 0, only the line is shown. If 1, line and bar is shown (if lineColor='', line is NOT shown)
//        lineColor=acolor    default is ''                Color of line (on top of histogram). Default is no line
//        class=aclass       default is class=tdHistoUp   Note that width and the 3 color values will override whatever is specified in class
// xlabel: array of x elements to note on x axis. values of this less than min or gt max are ignored

function histoPdf(anid,pobj,opts,xlabels,ylabels) {

var values,cdfs,minval,maxval,amean,mmax,m,aid,gotmean,psum,atir,t1,atd,adiv,thish,axx,acolor,foof2,foo3,a1,a2,aopt;
var mainColor='red',meanColor='lime', medianColor='red',myHeight='200',myClass='tdHistoUp',myWidth='10',myPadding='1';
var lineOnly=0,lineColor='',doClear=0 ;
var hiLiteObj={};
var hiLites=[],hiLiteColor='cyan';
var xlabelObj={};
var wasat,mrange,m,ival,ncols;
var ylabelUse=[],yUse;

var gotmean=0,psum=0, gotmedian=0;
if (arguments.length<4) xlabels=[];
if (arguments.length<5) ylabels=[];

if (arguments.length>2) {
  foof2=opts.replace(/\s+/g,','); foo3=foof2.split(',');
  for (m=0;m<foo3.length;m++) {
     a1=foo3[m];
     if (a1.indexOf('=')<0) continue ;
     a2=a1.split('=');
     aopt=a2[0].toUpperCase(); aval=a2[1];
     if (aopt=='COLOR') mainColor=aval;
     if (aopt=='MEANCOLOR') meanColor=aval;
     if (aopt=='MEDIANCOLOR') medianColor=aval;
     if (aopt=='HIGHLIGHTCOLOR') hiLiteColor=aval;
     if (aopt=='HILITECOLOR') hiLiteColor=aval;
     if (aopt=='HIGHLIGHT') hiLites=aval.split(';');
     if (aopt=='HILITE') hiLites=aval.split(';');
     if (aopt=='CLASS') myClass=aval;
     if (aopt=='HEIGHT') myHeight=aval;
     if (aopt=='WIDTH') myWidth=aval;
     if (aopt=='PADDING') myPadding=aval;
     if (aopt=='LINEONLY') lineOnly=aval;
     if (aopt=='CLEAR')  doClear=1;
     if (aopt=='LINECOLOR') lineColor=aval;
  }
}

values=pobj.pdf; cdfs=pobj.cdf; minval=pobj.min ; maxval=pobj.max ;

amean=-11111111111111 ;         // a hack (signals no pobj.mean)
if (typeof(pobj.mean)!=='undefined')   amean=parseInt(pobj.mean) ;

if (hiLites.length>0) {
  for (var im=0;im<hiLites.length;im++) {
       hiLiteObj[hiLites[im]]=1;
  }
}

wasat=minval-1;
mrange=maxval-minval;
for (var m=0;m<xlabels.length;m++) {
  ival=xlabels[m];
  if (isNaN(ival) && ival.indexOf('%')>-1) {            // convert % to location between min and max val
      ival=parseInt(ival);
      ival=parseInt(minval+ ival*(mrange/100));
      xlabels[m]=ival;
  }
  ncols=ival-wasat;
  xlabelObj[ival]=ncols;
  wasat=ival;
}

mmax=0;

for (var m=minval;m<=maxval;m++) {
  mmax=Math.max(values[m],mmax);
  if (values[m]<0) {
     alert('histoPdf error: bad value  at '+m+'='+values[m]);return null ;
  }
}

aid=document.getElementById(anid);
if (!aid) {
   alert("Error in histoPdf: no such target "+anid);
   return null ;
}
if (doClear==1) aid.innerHTML='';

t1=document.createElement('table');
t1.style.height=myHeight+'px';
t1.setAttribute('cellpadding',0);
t1.setAttribute('cellspacing',myPadding);

atr=document.createElement('tr'); atr.setAttribute('valign','bottom');
if (ylabels.length>0) {
  for (i=0;i<ylabels.length;i++) {
       ylabels[i]=parseInt(ylabels[i]);
       yUse=mmax*ylabels[i]/100; yUse=yUse.toPrecision(3) ;
       ylabelUse[i]=yUse;
  }

   atd=document.createElement('td');   atd.style.position='relative';
   adiv=document.createElement('div');
//adiv.style.backgroundColor='gray';
   adiv.style.height=myHeight+'px';
   adiv.style.fontSize='60%';
   adiv.style.position='relative';
   adiv.style.borderRightStyle='solid';
   adiv.style.borderRightColor='black';
   adiv.style.paddingLeft='45px';

   for (i1=ylabels.length-1;i1>=0;i1--) {
      aval=ylabels[i1];
      adiv1=document.createElement('div');
      d1=ylabelUse[i1];
      adiv1.innerHTML=d1 + '&nbsp;&#151;';
      adiv1.style.position='absolute';
      adiv1.style.top=Math.min(90,(100-aval))+'%';
      adiv1.style.left='3px';
      adiv1.style.float='left';
      adiv.appendChild(adiv1);
   }
    atd.appendChild(adiv);
    atr.appendChild(atd);

}                              // yaxis

for (var m=minval;m<=maxval;m++) {      // plot the values
   psum=psum+(m*values[m]);
   atd=document.createElement('td');
   adiv=document.createElement('div');

   adiv.class=myClass; adiv.setAttribute('class',myClass);

   thish=1+parseInt((myHeight-1)*values[m]/mmax);
   adiv.style.height=thish+'px' ;
   adiv.style.width=myWidth+'px' ;

   acolor=mainColor;

   axx=m;
   adiv.title='X='+axx+' w/pdf = '+values[m].toFixed(4);

   if (typeof(pobj.cdf)!=='undefined') {
     adiv.title=adiv.title+', cdf='+cdfs[m].toFixed(3);
     if (gotmedian==0 && cdfs[m]>0.50 ) {
        acolor=medianColor;  gotmedian=1;
     }
   }
   if (gotmean==0 && amean==m) {
       acolor=meanColor ; gotmean=1;
   }

   if (typeof(hiLiteObj[m])!=='undefined') acolor=hiLiteColor ;

   if (lineOnly=='1') acolor='white';
   adiv.style.backgroundColor=acolor ;
   if (lineColor!=='') {
        adiv.style.borderTopStyle='dotted';
        adiv.style.borderTopColor=lineColor;

   }

   adiv.innerHTML='&nbsp;';
   atd.appendChild(adiv);
   atr.appendChild(atd);
}

t1.appendChild(atr);

if (xlabels.length>0) {
   atr=document.createElement('tr');
   if (ylabels.length>0) {
      atd=document.createElement('td');
      adiv=document.createElement('span');
      adiv.innerHTML='';
      atd.appendChild(adiv);
      atr.appendChild(atd);
   }
   atd=document.createElement('td');
   atd.setAttribute('colspan',(1+maxval-minval));
   atd.style.borderTopStyle='solid';
   atd.style.borderTopColor='black';
   adiv=document.createElement('div'); adiv.innerHTML='';
   atd.appendChild(adiv);
   atr.appendChild(atd);
   t1.appendChild(atr);

   atr=document.createElement('tr'); atr.setAttribute('valign','bottom');

  if (ylabels.length>0) {
      atd=document.createElement('td');
      adiv=document.createElement('span');
      adiv.innerHTML='';
      atd.appendChild(adiv);
      atr.appendChild(atd);
  }


   for (var m=minval;m<=maxval;m++) {
     atd=document.createElement('td');
     adiv=document.createElement('div');   adiv.style.fontSize='40%';
     adiv.innerHTML='&nbsp;';
     if (typeof(xlabelObj[m])!=='undefined') adiv.style.backgroundColor='black';
     atd.appendChild(adiv);
     atr.appendChild(atd);
   }
   t1.appendChild(atr);

   atr=document.createElement('tr'); atr.setAttribute('valign','bottom');

  if (ylabels.length>0) {
      atd=document.createElement('td');
      adiv=document.createElement('span');
      adiv.innerHTML='';
      atd.appendChild(adiv);
      atr.appendChild(atd);
  }


   for (var m=0;m<xlabels.length;m++) {
     ival=parseInt(xlabels[m]); ncols=xlabelObj[ival];
     atd=document.createElement('td');  atd.setAttribute('colspan',ncols);     atd.setAttribute('align','right');
     adiv=document.createElement('span');     adiv.style.fontSize='60%';     adiv.innerHTML=ival;
     atd.appendChild(adiv);
     atr.appendChild(atd);
   }
   t1.appendChild(atr);

 }


 aid.appendChild(t1);
 return 0;
}

//============================== cdf amean
// change range of pdf
function changeRangePdf(pdfobj,newmin,newmax) {

var oldmin,oldmax,dold,dnew,dx,pdf1,cdf1,cold,cnew,ax,im,asum,amean,csum,iput,tval ;
 newmin=parseInt(newmin); newmax=parseInt(newmax);
 oldmin=pdfobj.min; oldmax=pdfobj.max;
 dold=oldmax-oldmin;
 dnew=newmax-newmin;
 dx=dold/dnew;
 pdf1=[]; cdf1=[];
 cold=0;

// goo='';
 for (im=0;im<=dnew;im++) {
    ax=oldmin+ ( im*dx);
    cnew=figCdfValue(ax,pdfobj);
    iput=im+newmin ;
    tval=Math.max(0,cnew-cold);  // catch wierd rounding errors
    pdf1[iput]=tval;
//    goo=goo+'\n im ax iput  '+im+' '+ax+ ' '+iput+' cnew ='+cnew+' value= '+pdf1[im+newmin];
    cold=cnew;
 }
// alert(goo);
// alert(pdf1);
 asum=0; for (var im=newmin;im<=newmax;im++) asum=asum+pdf1[im];
 csum=0;   amean=0;
 for (var im=newmin;im<=newmax;im++) {
    pdf1[im]=pdf1[im]/asum ;
    csum=csum+pdf1[im];
    cdf1[im]=csum;
    amean=(im*pdf1[im])+amean;
 }
 return {pdf:pdf1,cdf:cdf1,mean:amean,min:newmin,max:newmax} ;
}

//===============
// compute an array of weighted sums of several pdfs: each elment being the weighted sum of corresponding elements in the pdfs
//  return a pdf object; with 3 or 5 elements:
//     If donorm=1  pdf (normalized weighted sum), cdf, mean, min and max.
//        otherwise: pdf (the weighted sum), min and max
//    min and max max are the lowest min and hightest max of the pdfs
//  Note that the ranges of the several pdf objects do not have to overlap
//  This means that each (so an index in results is composed only from pdfs that cover this index value)
//  If donorm=1 (the default) return a proper "empirical pdf object" -- normalize so that sum across all indices=1, add a cdf and mean field are correct
//       otherwise, do not -- so pdf is NOT a proper pdf (it is a weight sum); and the mean and cdf fields = null
// Note that a non-normalized "pdf object" can be used as input to histoPdf.
//
//  pdfobjs: array of "empirical pdf objects"
//  weights: array of scalar weights
//  donorm: 1 to return as a pdf (normalizzed, etc). Otherwise, just return a weighted sum (in the pdf field)
//
//  returns a pdf object (if donorm!=1, cdf and mean elements are NOT in this pdf object)

function weightedSumPdf(pdfobjs,weights,donorm) {

var res,bmin,bmax,apdf,amin,amax,im2,im,cdf,amean,asum;

if (arguments.length<3) donorm='1';

if (pdfobjs.length!==weights.length) {
   alert('Error in weightedSumPdf: length of pdfobjs ('+pdfobjs.length+') ne lenfth of weights ('+weights.length);
   return null;
}
res=[];
bmin=1000000000;
bmax=-bmin ;

for (var im=0;im<pdfobjs.length;im++) {
  apdf=pdfobjs[im].pdf ;
  amin=pdfobjs[im].min;
  amax=pdfobjs[im].max;;
  bmin=Math.min(bmin,amin); bmax=Math.max(bmax,amax);
  aweight=weights[im];
  for (var im2=amin; im2<=amax;im2++) {
      if (typeof(res[im2])=='undefined') {
           res[im2]=aweight*apdf[im2] ;
      } else {
          res[im2]=(aweight*apdf[im2])+res[im2];
      }
   }
}

if (donorm=='1') {
    asum=0;
    for (var im=bmin;im<=bmax;im++) asum=asum+(res[im]);
    csum=0;    amean=0; cdf=[];
    for (var im=bmin;im<=bmax;im++) {
        res[im]=res[im]/asum;
        csum=csum+res[im];
        cdf[im]=csum;
        amean=amean+(res[im]*im);
    }
    return {cdf:cdf,mean:amean,pdf:res,min:bmin,max:bmax} ;
} else {
     return {pdf:res,min:bmin,max:bmax} ;
}

}